import os, seneca, time, redis, subprocess
from os import getenv as env
from seneca.engine.interface import SenecaInterface
from seneca.engine.interpreter import SenecaInterpreter
from seneca.constants.config import get_redis_port, get_redis_password, MASTER_DB, load_env

def cli(fn):
    def _fn(fname, *args, **kwargs):
        si = start_interface()
        print('''
################################################################################
    Operating on "{}"
################################################################################
        '''.format(fname))
        kwargs['scope'] = {
            'rt': {
                'author': kwargs.get('author'),
                'sender': kwargs.get('sender')
            }
        }
        try:
            with open(fname) as f:
                kwargs['fname'] = fname
                fn(si, f.read(), *args, **kwargs)
        except:
            fn(si, fname, *args, **kwargs)
        print('''
################################################################################
    Done.
################################################################################
        ''')
    return _fn

def start_interface():
    start_server()
    SenecaInterpreter.setup()
    return SenecaInterface()

def start_server():
    try:
        conn = redis.StrictRedis(host='localhost', port=get_redis_port(), db=MASTER_DB, password=get_redis_password())
        conn.ping()
    except Exception as ex:
        subprocess.run(['bash', './scripts/start.sh'])
        load_env()


@cli
def run_by_file(si, code_str, *args, **kwargs):
    fname = kwargs.get('fname')
    scope = kwargs.get('scope', {})
    si.execute_code_str(code_str, scope=scope)
    print('"{}" is RUN'.format(fname))

@cli
def publish(si, code_str, fullname, *args, **kwargs):
    scope = kwargs.get('scope', {})
    si.publish_code_str(fullname, code_str, keep_original=True, scope=scope)
    print('"{}" is PUBLISHED'.format(fullname))

@cli
def remove_code(si, fullname, *args, **kwargs):
    si.remove_code(fullname)
    print('"{}" is REMOVED'.format(fullname))

def main():
    import argparse, sys
    parser = argparse.ArgumentParser(description='Run your project on a docker bridge network')
    parser.add_argument('--run', '-r', type=str, help='Run the smart contract file', default=False)
    parser.add_argument('--publish', '-p', type=str, help='Publish the smart contract file')
    parser.add_argument('--name', '-n', type=str, help='Run Name of contract to publish or run')
    parser.add_argument('--author', '-a', type=str, help='Your name as the author of the contract', default="anonnymous_author")
    parser.add_argument('--sender', '-s', type=str, help='Your name as the sender of the contract', default="anonnymous_sender")
    parser.add_argument('--delete', '-d', type=str, help='Smart contract name to delete')
    args = parser.parse_args()

    if args.publish:
        assert args.name, 'Supply the --name option to name the contract you want to publish.'
        publish(args.publish, args.name, author=args.author, sender=args.sender)
    elif args.run:
        if args.run.endswith('.sen.py'):
            run_by_file(args.run, author=args.author, sender=args.sender)
    elif args.delete:
        remove_code(args.delete, author=args.author, sender=args.sender)
    else:
        parser.print_help(sys.stderr)



'''
Defines intermedite representation of mysql queries.
Representations are serializable and easy to analyze and augment.

* TODO: Type annoations
* TODO: have to be very careful with joins and '.' in fields, else people will
 potentially be able to write into other tabless
* TODO: consider using an abstract base class.
* TODO: replace asserts with raising custom exception type.
* TODO: figure out escaping/backticks
* TODO: Some basic joins?
* TODO: columnRows data type, fixed length, enter data positionally or dict, nice print
* TODO: Handle column name _count

* TODO: maybe do table exists
* TODO: order_by_desc
* TODO: get_table
* TODO: add order_by and limit to delete query

* TODO: Make API more consistent across inserts, updates, and upserts

NOTE: This module doesn't handle security, that must be done upstream.

NOTE: We don't enforce caller to have any criteria, i.e. without "WHERE" all
rows will be updated, this enforcement should be added in a higher level API to
prevent accidental updates of all records.
'''
from typing import Type, Dict, Tuple, List

#cls: Type[A]

from seneca.engine.storage.mysql_base import SQLType, get_py_to_sql_cast_func, cast_py_to_sql, escape_sql_pattern, TabularKVs, py_mysql_dict
from seneca.engine.util import *

class Query(object):
    pass


class QueryComponent(object):
    pass


### Query Parts ###
class ColumnDefinition(QueryComponent):
    '''Describes a column name, type, and unique constraint
    (Covered in other tests)
    '''
    @auto_set_fields
    def __init__(self, name, sql_type, unique=False):
        pass

    def to_sql(self):
        return intercalate(' ', [
          self.name,
          str(self.sql_type),
          'UNIQUE' if self.unique else None,
        ])


class AutoIncrementColumn(ColumnDefinition):
    '''A specific column definition for auto increment columns, intended for IDs.
    (Covered in other tests)
    '''
    @auto_set_fields
    def __init__(self, name: str):
        self.sql_type = 'BIGINT'# TODO: figure out if this should be INT or BIGINT
        self.is_unique = False

    def to_sql(self):
        return '%s %s unsigned NOT NULL AUTO_INCREMENT' % (self.name, self.sql_type)


class NonNullableBooleanColumn(ColumnDefinition):
    # Note: this is not intended for inclusion in end-user libs. Non-nullable
    # interferes with SPITS snapshotting.
    @auto_set_fields
    def __init__(self, name: str):
        self.sql_type = 'Boolean'
        self.is_unique = False

    def to_sql(self):
        return '%s %s NOT NULL DEFAULT FALSE' % (self.name, self.sql_type)


class QueryCriterion(QueryComponent):
    '''
    >>> QueryCriterion('eq', 'username', 'tester').to_sql()
    "username = 'tester'"

    >>> QueryCriterion('eq', 'username', None).to_sql()
    'username IS NULL'

    >>> QueryCriterion('ne', 'username', None).to_sql()
    'username IS NOT NULL'

    '''
    cr_types_to_strs = {
        'eq': '=',
        'ne': '!=',
        'lt': '<',
        'gt': '>',
        'le': '<=',
        'ge': '>=',
    }

    @auto_set_fields
    def __init__(self, constraint_type, field_name, comparison_value):
        assert self.constraint_type in list(self.cr_types_to_strs.keys()), "Invalid constraint type."

    def to_sql(self):
        if self.comparison_value is None:
            if self.constraint_type == 'eq':
                return "%s IS NULL" % self.field_name
            elif self.constraint_type == 'ne':
                return "%s IS NOT NULL" % self.field_name
            else:
                raise Exception("Unsupported operator with 'NONE' value only == (eq) and != (ne) are supported")
        else:
            return "%s %s %s" % (self.field_name,
                                 self.cr_types_to_strs[self.constraint_type],
                                 cast_py_to_sql(self.comparison_value),
                                )

class InvertedCriterion(QueryComponent):
    '''
    >>> InvertedCriterion(QueryCriterion('eq', 'username', 'tester')).to_sql()
    "(NOT username = 'tester')"
    '''
    @auto_set_fields
    def __init__(self, criterion):
        pass

    def to_sql(self):
        return "(NOT %s)" % self.criterion.to_sql()


class AndedCriteria(QueryComponent):
    '''
    >>> AndedCriteria(
    ...   [ QueryCriterion('eq', 'username', 'tester'),
    ...     QueryCriterion('gt', 'balance', 50)
    ... ]).to_sql()
    "(username = 'tester' AND balance > 50)"
    '''
    @auto_set_fields
    def __init__(self, criteria):
        pass

    def to_sql(self):
        return "(%s)" % ' AND '.join([x.to_sql() for x in self.criteria])


class OredCriteria(QueryComponent):
    '''
    >>> OredCriteria(
    ...   [ QueryCriterion('eq', 'username', 'tester'),
    ...     QueryCriterion('gt', 'balance', 50)
    ... ]).to_sql()
    "(username = 'tester' OR balance > 50)"
    '''
    @auto_set_fields
    def __init__(self, criteria):
        pass

    def to_sql(self):
        return "(%s)" % ' OR '.join([x.to_sql() for x in self.criteria])


# Function shared between InsertRows and SelectRows
def format_where_clause(crit):
    if crit:
        return 'WHERE %s' % crit.to_sql()
    else:
        return None

def make_order_by(by, desc):
    if by:
        ret = 'ORDER BY %s' % by
        if desc:
            return '%s DESC' % ret
        else:
            return ret
    else:
        assert desc is None, "Malformed order by, it only contains DESC value"
        return None

### Queries ###
class DeleteRows(Query):
    '''
    >>> print(DeleteRows('test_users', QueryCriterion('eq', 'username', 'test')).to_sql())
    DELETE FROM test_users
    WHERE username = 'test';
    '''
    @auto_set_fields
    def __init__(self, table_name, criteria, order_by=None, order_desc=None, limit=None):
        pass

    def to_sql(self):
        order_by_str = make_order_by(self.order_by, self.order_desc)

        return intercalate('\n',[
          'DELETE FROM %s' % self.table_name,
          'WHERE %s' % self.criteria.to_sql() if self.criteria else None,
          order_by_str,
          'LIMIT %d' % self.limit if self.limit else None,
        ]) + ';'


class UpdateRows(Query):
    '''
    >>> print(UpdateRows('test_users',
    ...   QueryCriterion('eq', 'username', 'tester'),
    ...   {'balance': 0, 'status':'broke'}
    ... ).to_sql())
    ...
    UPDATE test_users
    SET balance=0, status='broke'
    WHERE username = 'tester';
    '''
    @auto_set_fields
    def __init__(self, table_name, criteria, column_value_dict, order_by=None, order_desc=None, limit=None):
        # NOTE: doesn't really matter for this smart contracts, but orderby would be hel
        pass

    def to_sql(self):
        assignment_strs = ['%s=%s' % (k, cast_py_to_sql(v)) for (k,v) in self.column_value_dict.items()]
        order_by_str = make_order_by(self.order_by, self.order_desc)

        return intercalate('\n', [
          'UPDATE %s' % self.table_name,
          'SET %s' % intercalate(', ', assignment_strs),
          format_where_clause(self.criteria),
          order_by_str,
          'LIMIT %d' % self.limit if self.limit else None,
        ]) + ';'

# TODO: Change name to UpsertRows, see if this should actually be implemented with REPLACE
class SetRows(Query):
    # TODO: add unit test.
    # TODO: add query criterion and limit
    def __init__(self, table_name, list_of_values_lists):
        self.table_name = table_name
        # NOTE: Casting lists to tuples to ensure consistent types
        self.column_names = tuple(['k', 'v', 't'])
        self.list_of_values_lists = list(map(tuple, list_of_values_lists))

    def to_sql(self):
        correct_len = len(self.column_names)

        for idx, value_list in enumerate(self.list_of_values_lists):
            if len(self.list_of_values_lists[idx]) == correct_len - 1:
                self.list_of_values_lists[idx] = (*value_list, type(value_list[-1]).__name__)
            assert len(self.list_of_values_lists[idx]) == correct_len

        # NOTE: Given a list of rows, i.e. a list of lists, we take the the first
        # element of the list (a list representing a row). We use type info of
        # that first row to derrive type information
        casting_funcs = [get_py_to_sql_cast_func(type(x)) for x in self.list_of_values_lists[0]]

        def apply_funcs(xs):
            #''' Apply list of functions over a list of values '''
            #return [ f(x) for (f, x) in zip(casting_funcs, xs)]
            return map(cast_py_to_sql, xs)

        def format_string(xs):
            return '(%s)' % ', '.join(xs)

        def format_update(xs):
            return "v = %s" % list(xs)[1]

        convert_data_list_to_string = compose(format_string, apply_funcs)
        formatted_values_lists = [convert_data_list_to_string(x) for x in self.list_of_values_lists]
        convert_data_list_to_update = compose(format_update, apply_funcs)
        formatted_updates_lists = [convert_data_list_to_update(x) for x in self.list_of_values_lists]

        return '\n'.join([
          'INSERT INTO %s' % self.table_name,
          '(%s)' % ', '.join(self.column_names),
          'VALUES',
          ', '.join(formatted_values_lists),
          'ON DUPLICATE KEY UPDATE',
          ', '.join(formatted_updates_lists),
        ]) + ';'

class InsertRows(Query):
    '''
    >>> print(InsertRows('test_users', ['username', 'first_name', 'balance'],
    ...   [['tester', 'test', 500],
    ...    ['tester2', 'two', 200],
    ...   ]).to_sql())
    INSERT INTO test_users
    (username, first_name, balance)
    VALUES
    ('tester', 'test', 500), ('tester2', 'two', 200);
    '''
    def __init__(self, table_name, column_names, list_of_values_lists):
        self.table_name = table_name
        # NOTE: Casting lists to tuples to ensure consistent types
        self.column_names = tuple(column_names)
        self.list_of_values_lists = list(map(tuple, list_of_values_lists))


    def to_sql(self):
        correct_len = len(self.column_names)

        for value_list in self.list_of_values_lists:
            assert len(value_list) == correct_len

        # NOTE: Given a list of rows, i.e. a list of lists, we take the the first
        # element of the list (a list representing a row). We use type info of
        # that first row to derrive type information
        casting_funcs = [get_py_to_sql_cast_func(type(x)) for x in self.list_of_values_lists[0]]

        def apply_funcs(xs):
            #''' Apply list of functions over a list of values '''
            #return [ f(x) for (f, x) in zip(casting_funcs, xs)]
            return map(cast_py_to_sql, xs)

        def format_values(xs):
            return '(%s)' % ', '.join(xs)

        convert_data_list_to_string = compose(format_values, apply_funcs)
        formatted_values_lists = [convert_data_list_to_string(x) for x in self.list_of_values_lists]

        return '\n'.join([
          'INSERT INTO %s' % self.table_name,
          '(%s)' % ', '.join(self.column_names),
          'VALUES',
          ', '.join(formatted_values_lists),
        ]) + ';'


class SelectRows(Query):
    '''
    >>> print(SelectRows('test_users', [], None, None, None).to_sql())
    SELECT *
    FROM test_users;
    >>> print(SelectRows('test_users', [],
    ...  QueryCriterion('eq', 'username', 'tester'),
    ...  None,
    ...  None,
    ... ).to_sql())
    SELECT *
    FROM test_users
    WHERE username = 'tester';
    >>>
    >>> print(SelectRows('test_users', [],
    ...   QueryCriterion('gt', 'balance', 10),
    ...   None,
    ...   None,
    ... ).to_sql())
    SELECT *
    FROM test_users
    WHERE balance > 10;
    >>>
    >>> print(SelectRows('test_users', ['username', 'balance'],
    ...   QueryCriterion('gt', 'balance', 10),
    ...   None,
    ...   None,
    ... ).to_sql())
    SELECT username, balance
    FROM test_users
    WHERE balance > 10;
    >>>
    >>> print(SelectRows('test_users', ['username', 'balance'],
    ...   QueryCriterion('gt', 'balance', 10),
    ...   'balance',
    ...   None,
    ... ).to_sql())
    SELECT username, balance
    FROM test_users
    WHERE balance > 10
    ORDER BY balance;
    >>>
    >>> print(SelectRows('test_users', ['username', 'balance'],
    ...   QueryCriterion('gt', 'balance', 10),
    ...   limit=5,
    ... ).to_sql())
    SELECT username, balance
    FROM test_users
    WHERE balance > 10
    LIMIT 5;
    >>>
    >>> print(SelectRows('test_users', ['username', 'balance'],
    ...   QueryCriterion('gt', 'balance', 10),
    ...   'balance',
    ...   limit=5,
    ... ).to_sql())
    SELECT username, balance
    FROM test_users
    WHERE balance > 10
    ORDER BY balance
    LIMIT 5;
    '''

    # TODO: add orderby and limit
    @auto_set_fields
    def __init__(self,
                table_name,
                column_names,
                criteria,
                order_by=None,
                order_desc=None,
                limit=None):
        pass

    @staticmethod
    def format_column_names(c_names):
        if not c_names:
            # TODO: decide if we actually want this to auto-populate column names from python-side table object data instead
            return '*'
        else:
            return ', '.join(c_names)


    def to_sql(self):
        order_by_str = make_order_by(self.order_by, self.order_desc)

        return intercalate('\n', [
          'SELECT %s' % self.format_column_names(self.column_names),
          'FROM %s' % self.table_name,
          format_where_clause(self.criteria),
          order_by_str,
          'LIMIT %d' % self.limit if self.limit else None,
        ]) + ';'


class CountUniqueRows(Query):
    '''
    >>> print(CountUniqueRows('test_users', 'status', None).to_sql())
    SELECT status, COUNT(*) as _count
    FROM test_users
    GROUP BY status;

    >>> print(CountUniqueRows('test_users', 'status',
    ... QueryCriterion('eq', 'firstname', 'john')
    ... ).to_sql())
    SELECT status, COUNT(*) as _count
    FROM test_users
    WHERE firstname = 'john'
    GROUP BY status;
    '''
    @auto_set_fields
    def __init__(self, table_name, unique_column, criteria):
        pass

    def to_sql(self):
        return intercalate('\n', [
          'SELECT %s, COUNT(*) as _count' % self.unique_column,
          'FROM %s' % self.table_name,
          format_where_clause(self.criteria),
          'GROUP BY %s' % self.unique_column
        ]) + ';'


class CountRows(Query):
    '''
    >>> print(CountRows('test_users', None).to_sql())
    SELECT COUNT(*) as _count
    FROM test_users;

    >>> print(CountRows('test_users',
    ... QueryCriterion('eq', 'firstname', 'john')
    ... ).to_sql())
    SELECT COUNT(*) as _count
    FROM test_users
    WHERE firstname = 'john';
    '''
    @auto_set_fields
    def __init__(self, table_name, criteria):
        pass

    def to_sql(self):
        return intercalate('\n', [
          'SELECT COUNT(*) as _count',
          'FROM %s' % self.table_name,
          format_where_clause(self.criteria),
        ]) + ';'


class DescribeTable(Query):
    '''
    >>> print(DescribeTable('test_users').to_sql())
    DESCRIBE test_users;
    '''
    @auto_set_fields
    def __init__(self, table_name):
        pass

    def to_sql(self):
        return 'DESCRIBE %s;' % self.table_name


class ListTables(Query):
    '''
    >>> print(ListTables(None).to_sql())
    SHOW TABLES;
    >>> print(ListTables('tester_').to_sql())
    SHOW TABLES LIKE 'tester\_%;
    '''
    @auto_set_fields
    def __init__(self, prefix=None):
        pass

    @staticmethod
    def format_prefix(pf):
        if pf:
#            return 'LIKE \'%s\\_\%\'' % pf
            return 'LIKE \'{0}%'.format(escape_sql_pattern(pf))
        else:
            return None

    def to_sql(self):
        return intercalate(' ', [
            'SHOW TABLES',
            self.format_prefix(self.prefix),
        ]) + ';'



class AddTableColumn(Query):
    '''
    >>> print(AddTableColumn('test_users', ColumnDefinition('balance2', SQLType('BIGINT'), False)).to_sql())
    ALTER TABLE test_users
    ADD balance2 BIGINT;
    >>> print(AddTableColumn('test_users', ColumnDefinition('secondary_id', SQLType('VARCHAR', 30), True)).to_sql())
    ALTER TABLE test_users
    ADD secondary_id VARCHAR(30) UNIQUE;
    '''
    @auto_set_fields
    def __init__(self, table_name, column_def):
        pass

    def to_sql(self):
        #name, sql_type, is_unique
        return intercalate('\n',[
          'ALTER TABLE %s' % self.table_name,
          'ADD %s' % self.column_def.to_sql()
        ]) + ';'


class DropTableColumn(Query):
    '''
    >>> print(DropTableColumn('test_users', 'balance2').to_sql())
    ALTER TABLE test_users
    DROP COLUMN balance2;
    '''
    @auto_set_fields
    def __init__(self, table_name, column_name):
        pass

    def to_sql(self):
        return intercalate('\n', [
          'ALTER TABLE %s' % self.table_name,
          'DROP COLUMN %s' % self.column_name
        ]) + ';'


class DropTable(Query):
    '''
    >>> print(DropTable('test_users').to_sql())
    DROP TABLE test_users;
    '''
    @auto_set_fields
    def __init__(self, table_name):
        pass

    def to_sql(self):
        return 'DROP TABLE %s;' % self.table_name


class CreateTable(Query):
    '''
    >>> print(CreateTable(
    ...       'test_users',
    ...       AutoIncrementColumn('id'),
    ...       [ ColumnDefinition('username', SQLType('VARCHAR', 30), True),
    ...         ColumnDefinition('drivers_licence_unmber', SQLType('VARCHAR', 30), True),
    ...         ColumnDefinition('first_name', SQLType('VARCHAR', 30), False),
    ...         ColumnDefinition('balance', SQLType('BIGINT'), False),
    ... ]).to_sql())
    CREATE TABLE test_users (
    id BIGINT unsigned NOT NULL AUTO_INCREMENT PRIMARY KEY,
    username VARCHAR(30) UNIQUE,
    drivers_licence_unmber VARCHAR(30) UNIQUE,
    first_name VARCHAR(30),
    balance BIGINT
    );

    >>> print(CreateTable(
    ...       'test_users',
    ...       AutoIncrementColumn('id'),
    ...       [ ColumnDefinition('username', SQLType('VARCHAR', 30), True),
    ...         ColumnDefinition('drivers_licence_unmber', SQLType('VARCHAR', 30), True),
    ...         ColumnDefinition('first_name', SQLType('VARCHAR', 30), False),
    ...         ColumnDefinition('balance', SQLType('BIGINT'), False),
    ... ], if_not_exists=True).to_sql())
    CREATE TABLE IF NOT EXISTS test_users (
    id BIGINT unsigned NOT NULL AUTO_INCREMENT PRIMARY KEY,
    username VARCHAR(30) UNIQUE,
    drivers_licence_unmber VARCHAR(30) UNIQUE,
    first_name VARCHAR(30),
    balance BIGINT
    );
    '''
    @auto_set_fields
    def __init__(self, table_name, primary_key_column_def, other_column_defs, if_not_exists=False):
        #print(self.__dict__)
        pass

    def to_sql(self):
        column_def_strs = (
          [self.primary_key_column_def.to_sql() + ' PRIMARY KEY'] +
          list(map(lambda x: x.to_sql(), self.other_column_defs))
        )

        ine = 'IF NOT EXISTS' if self.if_not_exists else ''

        return '\n'.join([
          'CREATE TABLE %s (' % intercalate(' ', [ine, self.table_name]),
          intercalate(',\n', column_def_strs),
          ');'
        ])


def run_tests(_):
    '''
    >>> NonNullableBooleanColumn('test').to_sql()
    'test Boolean NOT NULL DEFAULT FALSE'
    '''
    import doctest, sys
    return doctest.testmod(sys.modules[__name__])



from seneca.engine.storage.redisnap.commands import *
#import seneca.engine.storage.redisnap.resp_types as rtype
#from seneca.engine.storage.redisnap.addresses import *

#https://stackoverflow.com/questions/40508492/python-sphinx-inherit-method-documentation-from-superclass/40613306#40613306
import abc

class SuperclassMeta(type):
    def __new__(mcls, classname, bases, cls_dict):
        cls = super().__new__(mcls, classname, bases, cls_dict)
        for name, member in cls_dict.items():
            try:
                if not getattr(member, '__doc__'):
                    member.__doc__ = getattr(bases[-1], name).__doc__
            except:
                AttributeError
        return cls


class ExecuterBase(object, metaclass=SuperclassMeta):
    '''
    Maps command objects to actual Redis commands and runs them, leans heavily
    on redis.py

    TODO:
    We should efficiently track collisions and decide whether we want to
    use a log of transactions to commit, or create ops from the stored data
    '''

    @abc.abstractmethod
    def purge(self):
        """
        """


    @abc.abstractmethod
    def exists(self, cmd):
        """
        >>> _ = ex.purge()
        >>> ex(Exists('foo'));
        False
        """


    @abc.abstractmethod
    def type(self, cmd):
        """
        This is the built-in shallow Redis typecheck.

        >>> _ = ex.purge()
        >>> t = ex(Type('foo'))
        >>> print(t.__name__)
        RDoesNotExist
        >>> issubclass(t, RScalar)
        True
        """


    @abc.abstractmethod
    def asserttype(self, cmd):
        """
        >>> ex.purge()
        >>> ex(SetNR('foo', 'bar'))
        >>> ex(AssertType('foo', RScalar))
        True
        >>> ex(AssertType('foo', RHash))
        False

        NOTE: This is the exact same implementation in local and redis backends
        """

    # # TODO:
    # @abc.abstractmethod
    # def assertfieldtype(self, cmd):
    #     """
    #     This is not part of RESP, it's a RediSnap add-on, it does deep type
    #     inspection.
    #
    #     NOTE: Very important, In Redis empty fields are fully polymorphic, so
    #     asserting any type on non-existent key will always succeed. This is how
    #     Redis behaves:
    #     > get some_non_existent_key
    #     (nil)
    #
    #     Only if an existing value exists of the wrong type is there a problem:
    #     > set some_string abc
    #     OK
    #     > hget some_string field_name
    #     (error) WRONGTYPE Operation against a key holding the wrong kind of value
    #
    #     TODO: Add a test for this command
    #     >> ex(AssertFieldType('foo', 'bar', RScalar))
    #     """

    @abc.abstractmethod
    def get(self, cmd):
        """
        >>> _ = ex.purge()
        >>> ex(Get('foo'))
        RDoesNotExist()
        """


    @abc.abstractmethod
    def setnr(self, cmd):
        """
        >>> _ = ex.purge()
        >>> ex(SetNR('foo', 'bar'))

        >>> ex(Exists('foo'))
        True

        >>> ex(Type('foo')).__name__; ex(Get('foo'))
        'RScalar'
        RScalar('bar')

        >>> _ = ex(SetNR('foo', 1)); ex(Type('foo')).__name__; ex(Get('foo'))
        'RScalar'
        RScalarInt(1)
        """


    @abc.abstractmethod
    def incrbynr(self, cmd):
        """
        >>> ex.purge()

        Increment an empty key
        >>> ex(IncrByNR('foo', 1));

        >>> ex(Get('foo'))
        RScalarInt(1)

        Increment an existing key
        >>> ex(IncrByNR('foo', 1));

        >>> ex(Get('foo'))
        RScalarInt(2)

        Incremenent non-int scalars
        >>> ex(SetNR('foo', 'bar'))
        >>> return_exception_tuple(ex, IncrByNR('foo', 1))
        ('RedisVauleTypeError', 'Existing value has wrong type.')

        >>> ex(SetNR('foo', 1.0))
        >>> return_exception_tuple(ex, IncrByNR('foo', 1))
        ('RedisVauleTypeError', 'Existing value has wrong type.')
        """


    @abc.abstractmethod
    def appendnr(self, cmd):
        """
        >>> ex.purge()
        >>> ex(AppendNR('foo', 'abc')); ex(Get('foo'))
        RScalar('abc')
        >>> ex(AppendNR('foo', 'abc')); ex(Get('foo'))
        RScalar('abcabc')

        >>> ex(AppendNR('fooint', '1')); ex(Get('fooint'))
        RScalarInt(1)
        >>> ex(AppendNR('fooint', '1')); ex(Get('fooint'))
        RScalarInt(11)
        """


    @abc.abstractmethod
    def hget(self, cmd):
        """
        >>> ex.purge()
        >>> ex(HGet('foo', 'bar'))
        RDoesNotExist()
        """


    @abc.abstractmethod
    def hsetnr(self, cmd):
        """
        >>> ex.purge()
        >>> ex(HSetNR('foo', 'bar', 'baz'))
        >>> ex(HGet('foo', 'bar'))
        RScalar('baz')

        >>> ex(HSetNR('foo', 'bar', 1))
        >>> ex(HGet('foo', 'bar'))
        RScalarInt(1)
        """


    @abc.abstractmethod
    def hexists(self, cmd):
        """
        >>> ex.purge()
        >>> ex(HExists('foo', 'bar'))
        False

        >>> ex(HSetNR('foo', 'bar', 'baz'))

        >>> ex(HExists('foo', 'qux'))
        False
        >>> ex(HExists('foo', 'bar'))
        True

        >>> ex(SetNR('foo', 'bar'))
        >>> return_exception_tuple(ex, HExists('foo', 'bar'))
        ('RedisKeyTypeError', 'Existing value has wrong type.')
        """


    @abc.abstractmethod
    def del_(self, cmd):
        """
        """


    # # TODO:
    # @abc.abstractmethod
    # def llen(self, cmd):
    #     """
    #     TODO: add test for llen
    #     >> ex(LLen('foo'))
    #     """


    @abc.abstractmethod
    def lindex(self, cmd):
        """
        >>> ex.purge()

        >>> ex(RPushNR('foo', ['bar']))
        >>> ex(LIndex('foo', 0))
        RScalar('bar')

        >>> ex(LIndex('foo', 20))
        RDoesNotExist()

        >>> ex.purge()

        >>> ex(LIndex('foo', 0))
        RDoesNotExist()

        >>> ex(SetNR('foo', 'bar'))
        >>> return_exception_tuple(ex, LIndex('foo', 0))
        ('RedisKeyTypeError', 'Existing value has wrong type.')
        """


    @abc.abstractmethod
    def lset(self, cmd):
        """
        >>> ex.purge()
        >>> return_exception_tuple(ex, LSet('foo', 0, 'bar'))
        ('RedisKeyTypeError', 'Cannot LSet an nonexistent key.')

        >>> ex(RPushNR('foo', ['bar']))
        >>> ex(LSet('foo', 0, 'baz'))

        >>> return_exception_tuple(ex, LSet('foo', 1, 'bar'))
        ('RedisListOutOfRange', 'Index out of range.')
        """


    @abc.abstractmethod
    def lpushnr(self, cmd):
        """
        >>> ex.purge()
        >>> ex(LPushNR('foo', ['bar']))
        >>> ex(LIndex('foo', 0))
        RScalar('bar')

        >>> ex(LPushNR('foo', ['baz']))
        >>> ex(LIndex('foo', 0)); ex(LIndex('foo', 1));
        RScalar('baz')
        RScalar('bar')

        >>> ex.purge()
        >>> ex(LPushNR('foo', ['bar', 'baz']))
        >>> ex(LIndex('foo', 0)); ex(LIndex('foo', 1));
        RScalar('baz')
        RScalar('bar')
        """


    @abc.abstractmethod
    def rpushnr(self, cmd):
        """
        >>> ex.purge()
        >>> ex(RPushNR('foo', ['bar']))
        >>> ex(LIndex('foo', 0))
        RScalar('bar')

        >>> ex(RPushNR('foo', ['baz']))
        >>> ex(LIndex('foo', 0)); ex(LIndex('foo', 1));
        RScalar('bar')
        RScalar('baz')


        >>> ex.purge()
        >>> ex(RPushNR('foo', ['bar', 'baz']))
        >>> ex(LIndex('foo', 0)); ex(LIndex('foo', 1));
        RScalar('bar')
        RScalar('baz')
        """


    @abc.abstractmethod
    def lpop(self, cmd):
        """
        >>> ex.purge()
        >>> ex(LPop('foo'))
        RDoesNotExist()

        >>> ex(SetNR('foo', 'bar'))
        >>> return_exception_tuple(ex, LPop('foo'))
        ('RedisKeyTypeError', 'Existing value has wrong type.')

        >>> ex.purge()
        >>> ex(LPushNR('foo', ['bar', 'baz']))
        >>> ex(LPop('foo')); ex(LPop('foo'))
        RScalar('baz')
        RScalar('bar')
        """


    @abc.abstractmethod
    def rpop(self, cmd):
        """
        >>> ex.purge()
        >>> ex(RPushNR('foo', ['bar', 'baz']))
        >>> ex(RPop('foo')); ex(RPop('foo'))
        RScalar('baz')
        RScalar('bar')
        """


    @abc.abstractmethod
    def zaddnr(self, cmd):
        """
        >>> ex.purge()

        # Testing auto creation
        >>> ex(ZAddNR('foo', {'bar':1}))
        >>> ex(ZScore('foo', 'bar'))
        1

        # Testing modification of an existing value
        >>> ex(ZAddNR('foo', {'baz':2}))
        >>> ex(ZScore('foo', 'bar'));  ex(ZScore('foo', 'baz'))
        1
        2

        # Testing update of an existing member
        >>> ex(ZAddNR('foo', {'baz':50}))
        >>> ex(ZScore('foo', 'bar')); ex(ZScore('foo', 'baz'))
        1
        50

        # Testing exception on type mismatch
        >>> ex(SetNR('foo', 'bar'))
        >>> return_exception_tuple(ex, ZAddNR('foo', {'baz':2}))
        ('RedisKeyTypeError', 'Existing value has wrong type.')
        """


    @abc.abstractmethod
    def zscore(self, cmd):
        """
        >> ex.purge()

        Empty key returns None
        >> ex(ZScore('foo', 'bar'))

        Member is not present returns None too
        >> ex(ZAddNR('foo', {'bar':1})); ex(ZScore('foo', 'baz'))

        >> ex(ZScore('foo', 'bar'))
        1

        # Testing exception on type mismatch
        >> ex(SetNR('foo', 'bar'))
        >> return_exception_tuple(ex, ZScore('foo', 'bar'))
        ('RedisKeyTypeError', 'Existing value has wrong type.')
        """


    @abc.abstractmethod
    def zrevrangebyscore(self, cmd):
        """
        Removes the specified members from the sorted set stored at key. Non existing members are ignored.
        >>> ex.purge()
        >>> ex(ZAddNR('foo', {'one': 1, 'two':2, 'three':3}))
        >>> list(ex(ZRevRangeByScore('foo',None,None)))
        ['three', 'two', 'one']

        >>> list(ex(ZRevRangeByScore('foo', 2, 1)))
        ['two', 'one']

        >>> list(ex(ZRevRangeByScore('foo', 2, 1, (True, False))))
        ['two']

        >>> list(ex(ZRevRangeByScore('foo', 2, 1, (False, False))))
        []

        >>> list(ex(ZRevRangeByScore('foo', None, None, with_scores=True)))
        [(3, 'three'), (2, 'two'), (1, 'one')]

        # Testing exception on type mismatch
        >>> ex(SetNR('foo', 'bar'))
        >>> return_exception_tuple(ex, ZRevRangeByScore('foo',10,30))
        ('RedisKeyTypeError', 'Existing value has wrong type.')
        """


    @abc.abstractmethod
    def zremnr(self, cmd):
        """
        >>> ex.purge()

        Testing zrem on non existing key
        >>> ex(ZRemNR('foo', ['bar']))
        >>> ex(Exists('foo'))
        False

        Test modification of existing
        >>> ex(ZAddNR('foo', {'bar':1, 'baz':2}))
        >>> ex(ZRemNR('foo', ['baz']))
        >>> ex(ZScore('foo', 'bar')); ex(ZScore('foo', 'baz'))
        1

        Test zrem of non-existent member
        >>> ex(ZRemNR('foo', ['bar']))

        Test deletion of sset after last member is removed
        >>> ex(Exists('foo'))
        False

        # Testing exception on type mismatch
        >>> ex(SetNR('foo', 'bar'))
        >>> return_exception_tuple(ex, ZRemNR('foo', ['qux']))
        ('RedisKeyTypeError', 'Existing value has wrong type.')


        # Test multiple removals in single invocation
        >>> ex(Del('foo'))
        >>> ex(ZAddNR('foo', {'bar':1, 'baz':2}))
        >>> ex(ZRemNR('foo', ['bar', 'baz']))

        >>> ex(Exists('foo'))
        False
        """


    @abc.abstractmethod
    def zincrbynr(self, cmd):
        """
        >>> ex.purge()

        # Totally empty key
        >>> ex(ZIncrByNR('foo', 1, 'bar')); ex(ZScore('foo', 'bar'))
        1

        # Missing member
        >>> ex(ZIncrByNR('foo', 1, 'baz')); ex(ZScore('foo', 'baz'))
        1

        # Update existing
        >>> ex(ZIncrByNR('foo', 4, 'baz')); ex(ZScore('foo', 'baz'))
        5
        """

    def __call__(self, cmd):
        # TODO: pass Make sure this is efficient and generally okay.

        if isinstance(cmd, Del):
            self.del_(cmd)
        # TODO: pass fix name of set
        #elif isinstance(cmd, Sel): pass
        #    self.sel_(cmd)
        else:
            return getattr(self, cmd.__class__.__name__.lower())(cmd)


def run_tests(deps_provider):
    from seneca.engine.util import return_exception_tuple
    from collections import namedtuple
    #import doctest, sys
    #return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})
    d = {'failed':0, 'attempted':0}
    return namedtuple('_', ' '.join(d.keys()))(**d)



from unittest import TestCase
from unittest.mock import MagicMock
from seneca.engine.client import *
from seneca.engine.interface import SenecaInterface
from seneca.engine.interpreter import SenecaInterpreter
from seneca.libs.logger import overwrite_logger_level


GENESIS_AUTHOR = 'davis'


XFER_CODE_STR = """ \

from seneca.contracts.currency import transfer
transfer('{receiver}', {amount})
"""


MINT_CODE_STR = """ \

from seneca.contracts.currency import mint
mint('davis', 10000)
mint('stu', 69)
mint('birb', 8000)
mint('ghu', 9000)
mint('tj', 8000)
mint('ethan', 8000)
"""


class MockContract:
    def __init__(self, sender: str, code: str, contract_name: str):
        self.sender, self.code, self.contract_name = sender, code, contract_name


def create_currency_tx(sender: str, receiver: str, amount: int, contract_name: str='currency'):
    code = XFER_CODE_STR.format(receiver=receiver, amount=amount)
    contract = MockContract(sender=sender, code=code, contract_name=contract_name)
    return contract


class TestSenecaClient(TestCase):
    CONTRACTS_TO_STORE = {'currency': 'kv_currency.sen.py'}

    def assert_completion(self, expected_sbb_rep: List[tuple]=None, input_hash=''):
        def _completion_handler(cr_data: CRContext):
            if input_hash:
                self.assertEqual(cr_data.input_hash, input_hash)
            if expected_sbb_rep:
                self.assertEqual(expected_sbb_rep, cr_data.get_subblock_rep())

        return _completion_handler

    def setUp(self):
        # overwrite_logger_level(0)
        with SenecaInterface(False) as interface:
            interface.r.flushall()
            # Store all smart contracts in CONTRACTS_TO_STORE
            import seneca
            test_contracts_path = seneca.__path__[0] + '/../test_contracts/'

            for contract_name, file_name in self.CONTRACTS_TO_STORE.items():
                with open(test_contracts_path + file_name) as f:
                    code_str = f.read()
                    interface.publish_code_str(contract_name, GENESIS_AUTHOR, code_str, keep_original=True)

            rt = {
                'author': GENESIS_AUTHOR,
                'sender': GENESIS_AUTHOR,
                'contract': 'minter'
            }
            interface.execute_code_str(MINT_CODE_STR, scope={'rt': rt})

    def test_setup_dbs(self):
        client = SenecaClient(sbb_idx=0, num_sbb=1)

        self.assertTrue(client.master_db is not None)
        self.assertTrue(client.active_db is None)

        self.assertEqual(len(client.available_dbs), NUM_CACHES)

    def test_flush(self):
        client = SenecaClient(sbb_idx=0, num_sbb=1)

        c1 = create_currency_tx('davis', 'stu', 14)
        c2 = create_currency_tx('stu', 'davis', 40)

        client._start_sb('A' * 64)
        client.run_contract(c1)
        client.run_contract(c2)

        client.flush_all()

        self.assertEqual(len(client.pending_dbs), 0)
        self.assertEqual(client.active_db, None)

    def test_run_tx_increments_contract_idx(self):
        client = SenecaClient(sbb_idx=0, num_sbb=1)
        client._start_sb('A' * 64)

        self.assertEqual(client.active_db.next_contract_idx, 0)

        c1 = create_currency_tx('davis', 'stu', 14)
        c2 = create_currency_tx('stu', 'davis', 40)

        client.run_contract(c1)
        self.assertEqual(client.active_db.next_contract_idx, 1)

        client.run_contract(c2)
        self.assertEqual(client.active_db.next_contract_idx, 2)

    def test_end_subblock_1_sbb(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        input_hash = 'A' * 64
        c1 = create_currency_tx('davis', 'stu', 14)
        c2 = create_currency_tx('stu', 'davis', 40)
        expected_sbb_rep = [(c1, "SUCC", "SET balances:davis 9986;SET balances:stu 83;"),
                            (c2, "SUCC", "SET balances:stu 43;SET balances:davis 10026;")]

        client = SenecaClient(sbb_idx=0, num_sbb=1, loop=loop)
        client._start_sb(input_hash)

        client.run_contract(c1)
        client.run_contract(c2)

        client._end_sb(self.assert_completion(expected_sbb_rep, input_hash))
        self.assertTrue(input_hash in client.pending_futures)

        # We must run the future manually, since the event loop is not currently running
        loop.run_until_complete(client.pending_futures[input_hash]['fut'])

        loop.close()

    def test_execute_sb(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        input_hash = 'A' * 64
        c1 = create_currency_tx('davis', 'stu', 14)
        c2 = create_currency_tx('stu', 'davis', 40)
        expected_sbb_rep = [(c1, "SUCC", "SET balances:davis 9986;SET balances:stu 83;"),
                            (c2, "SUCC", "SET balances:stu 43;SET balances:davis 10026;")]

        client = SenecaClient(sbb_idx=0, num_sbb=1, loop=loop)

        client.execute_sb(input_hash=input_hash, contracts=[c1, c2],
                          completion_handler=self.assert_completion(expected_sbb_rep, input_hash))
        self.assertTrue(input_hash in client.pending_futures)

        # We must run the future manually, since the event loop is not currently running
        loop.run_until_complete(client.pending_futures[input_hash]['fut'])

        loop.close()

    def test_end_subblock_1_sbb_with_failure(self):

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        input_hash = 'A' * 64
        c1 = create_currency_tx('davis', 'stu', 14)
        c2 = create_currency_tx('stu', 'davis', 9000)
        c3 = create_currency_tx('stu', 'davis', 40)
        expected_sbb_rep = [(c1, "SUCC", "SET balances:davis 9986;SET balances:stu 83;"),
                            (c2, "FAIL -- Sender balance must be non-negative!!!", ""),
                            (c3, "SUCC", "SET balances:stu 43;SET balances:davis 10026;")]

        client = SenecaClient(sbb_idx=0, num_sbb=1, loop=loop)
        client._start_sb(input_hash)

        client.run_contract(c1)
        client.run_contract(c2)
        client.run_contract(c3)

        client._end_sb(self.assert_completion(expected_sbb_rep, input_hash))
        self.assertTrue(input_hash in client.pending_futures)

        # We must run the future manually, since the event loop is not currently running
        loop.run_until_complete(client.pending_futures[input_hash]['fut'])

        loop.close()

    def test_end_subblock_2_sbb(self):
        input_hash1 = 'A' * 64
        input_hash2 = 'B' * 64

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        c1 = create_currency_tx('davis', 'stu', 14)
        c2 = create_currency_tx('stu', 'davis', 40)
        c3 = create_currency_tx('ghu', 'davis', 15)
        c4 = create_currency_tx('tj', 'birb', 90)
        expected_sbb1_rep = [(c1, "SUCC", "SET balances:davis 9986;SET balances:stu 83;"),
                             (c2, "SUCC", "SET balances:stu 43;SET balances:davis 10026;")]
        expected_sbb2_rep = [(c3, "SUCC", "SET balances:ghu 8985;SET balances:davis 10041;"),
                             (c4, "SUCC", "SET balances:tj 7910;SET balances:birb 8090;")]

        client1 = SenecaClient(sbb_idx=0, num_sbb=2, loop=loop)
        client2 = SenecaClient(sbb_idx=1, num_sbb=2, loop=loop)
        client1._start_sb(input_hash1)
        client2._start_sb(input_hash2)

        client1.run_contract(c1)
        client1.run_contract(c2)
        client2.run_contract(c3)
        client2.run_contract(c4)

        client1._end_sb(self.assert_completion(expected_sbb1_rep, input_hash1))
        client2._end_sb(self.assert_completion(expected_sbb2_rep, input_hash2))
        self.assertTrue(input_hash1 in client1.pending_futures)
        self.assertTrue(input_hash2 in client2.pending_futures)

        # We must run the future manually, since the event loop is not currently running
        coros = (client1.pending_futures[input_hash1]['fut'], client2.pending_futures[input_hash2]['fut'])
        loop.run_until_complete(asyncio.gather(*coros))

        loop.close()

    def test_end_subblock_2_sbb_start_subblocks_before_ending_then_flush_all_dem_bois(self):
        input_hash1 = 'A' * 64
        input_hash2 = 'B' * 64
        input_hash3 = 'C' * 64
        input_hash4 = 'D' * 64

        c1 = create_currency_tx('davis', 'stu', 14)
        c2 = create_currency_tx('stu', 'davis', 40)
        c3 = create_currency_tx('ghu', 'davis', 15)
        c4 = create_currency_tx('tj', 'birb', 90)
        c5 = create_currency_tx('ethan', 'birb', 60)
        c6 = create_currency_tx('stu', 'davis', 10)
        c7 = create_currency_tx('ghu', 'tj', 50)
        c8 = create_currency_tx('birb', 'davis', 100)
        expected_sbb1_1 = [(c1, "SUCC", "SET balances:davis 9986;SET balances:stu 83;"),
                           (c2, "SUCC", "SET balances:stu 43;SET balances:davis 10026;")]
        expected_sbb2_1 = [(c3, "SUCC", "SET balances:ghu 8985;SET balances:davis 10041;"),
                           (c4, "SUCC", "SET balances:tj 7910;SET balances:birb 8090;")]
        expected_sbb1_2 = [(c5, "SUCC", "SET balances:ethan 7940;SET balances:birb 8150;"),
                           (c6, "SUCC", "SET balances:stu 33;SET balances:davis 10051;")]
        expected_sbb2_2 = [(c7, "SUCC", "SET balances:ghu 8935;SET balances:tj 7960;"),
                           (c8, "SUCC", "SET balances:birb 8050;SET balances:davis 10151;")]

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        client1 = SenecaClient(sbb_idx=0, num_sbb=2, loop=loop)
        client2 = SenecaClient(sbb_idx=1, num_sbb=2, loop=loop)

        client1._start_sb(input_hash1)
        client2._start_sb(input_hash2)

        client1.run_contract(c1)
        client1.run_contract(c2)
        client2.run_contract(c3)
        client2.run_contract(c4)

        client1._end_sb(self.assert_completion(expected_sbb1_1, input_hash1))
        client2._end_sb(self.assert_completion(expected_sbb2_1, input_hash2))

        client1._start_sb(input_hash3)
        client2._start_sb(input_hash4)

        client1.run_contract(c5)
        client1.run_contract(c6)
        client2.run_contract(c7)
        client2.run_contract(c8)

        client1._end_sb(self.assert_completion(expected_sbb1_2, input_hash3))
        client2._end_sb(self.assert_completion(expected_sbb2_2, input_hash4))

        self.assertTrue(input_hash1 in client1.pending_futures)
        self.assertTrue(input_hash3 in client1.pending_futures)
        self.assertTrue(input_hash2 in client2.pending_futures)
        self.assertTrue(input_hash4 in client2.pending_futures)

        client1.update_master_db(input_hash1)
        client2.update_master_db(input_hash2)

        # We must run the future manually, since the event loop is not currently running
        coros = (client1.pending_futures[input_hash1]['fut'], client2.pending_futures[input_hash2]['fut'])
        loop.run_until_complete(asyncio.gather(*coros))

        coros = (client1.pending_futures[input_hash3]['fut'], client2.pending_futures[input_hash4]['fut'])
        loop.run_until_complete(asyncio.gather(*coros))

        loop.close()

    def test_update_master_db_with_incomplete_sb(self):
        input_hash1 = 'A' * 64
        input_hash2 = 'B' * 64
        input_hash3 = 'C' * 64
        input_hash4 = 'D' * 64

        c1 = create_currency_tx('davis', 'stu', 14)
        c2 = create_currency_tx('stu', 'davis', 40)
        c3 = create_currency_tx('ghu', 'davis', 15)
        c4 = create_currency_tx('tj', 'birb', 90)
        c5 = create_currency_tx('ethan', 'birb', 60)
        c6 = create_currency_tx('stu', 'davis', 10)
        c7 = create_currency_tx('ghu', 'tj', 50)
        c8 = create_currency_tx('birb', 'davis', 100)

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        client1 = SenecaClient(sbb_idx=0, num_sbb=2, loop=loop)
        client2 = SenecaClient(sbb_idx=1, num_sbb=2, loop=loop)

        client1._start_sb(input_hash1)
        client2._start_sb(input_hash2)

        client1.run_contract(c1)
        client1.run_contract(c2)
        client2.run_contract(c3)
        client2.run_contract(c4)

        client1._end_sb(self.assert_completion(None, input_hash1))
        client2._end_sb(self.assert_completion(None, input_hash2))

        client1._start_sb(input_hash3)
        client2._start_sb(input_hash4)

        client1.run_contract(c5)
        client1.run_contract(c6)
        client2.run_contract(c7)
        client2.run_contract(c8)

        client1._end_sb(self.assert_completion(None, input_hash3))
        client2._end_sb(self.assert_completion(None, input_hash4))

        client1.update_master_db(input_hash1)
        client2.update_master_db(input_hash2)
        client1.update_master_db(input_hash3)
        client2.update_master_db(input_hash4)

        self.assertTrue(input_hash1 in client1.pending_futures)
        self.assertTrue(input_hash3 in client1.pending_futures)
        self.assertTrue(input_hash2 in client2.pending_futures)
        self.assertTrue(input_hash4 in client2.pending_futures)

        # We must run the future manually, since the event loop is not currently running
        coros = (client1.pending_futures[input_hash1]['fut'], client2.pending_futures[input_hash2]['fut'],
                 client1.pending_futures[input_hash3]['fut'], client2.pending_futures[input_hash4]['fut'])
        loop.run_until_complete(asyncio.gather(*coros))

        for c in (client1, client2):
            self.assertEqual(len(c.pending_dbs), 0)

        loop.close()

    # Test that pending_db/active_db/working_db get updated as we go thru the flow

    # Test starting a new sub block before the last sub block finishes

    # Test with multiple sb's where stuff in SB 2 will pass the first time and fail the second time (cause some og read was modified)



'''
MySQL query execution and results parsing

TODOs:
  * Replace asserts with custom exceptions
  * Refactor move exception handling into dedicated code blocks
  * Catch specific mysql warnings that we don't care about
  * Create exceptions for:
    * Table already exists
    * What else?
  * Make sure we're using cursor correctly
  * Construct executer from existing connection (mainly for testing)
'''
import MySQLdb
from seneca.engine.storage.mysql_intermediate import *
import warnings

requires_commit = [ DeleteRows,
                    UpdateRows,
                    SetRows,
                    InsertRows,
                    AddTableColumn,
                    DropTableColumn,
                    DropTable,
                    CreateTable,
]

class SQLExecutionResult(object):
    @auto_set_fields
    def __init__(self, success, data):
        pass
    def __str__(self):
        return 'SQLExecutionResult(%s)' % str(self.__dict__)


def handle_error(q_type, err):
    return SQLExecutionResult(False, {'error_code': err.args[0],
                                      'error_message': err.args[1]})


def format_result(q_type, cur):
    # TODO: change to make_format_results_function

    def full_data_results(cur):
        keys = [x[0] for x in cur.description]
        data = TabularKVs(keys, cur.fetchall())
        return SQLExecutionResult(True, data)

    def simple_results(cur):
        return SQLExecutionResult(True, None)

    def not_implemented(cur):
        raise Exception('Not implemented')

    def id_count_results(cur):
        return SQLExecutionResult(True,
            { 'last_row_id':cur.lastrowid,
              'row_count':cur.rowcount
            })

    def col_1_row_1_results(cur):
        return SQLExecutionResult(True,cur.fetchone()[0])

    def col_1_values_results(cur):
        return SQLExecutionResult(True, [x[0] for x in cur.fetchall()])

    parser_funcs = { CreateTable: simple_results,
                     DescribeTable: full_data_results,
                     AddTableColumn: simple_results,
                     DropTableColumn: simple_results,
                     InsertRows: id_count_results,
                     UpdateRows: id_count_results,
                     SetRows: id_count_results,
                     SelectRows: full_data_results,
                     CountUniqueRows: full_data_results,
                     CountRows: col_1_row_1_results,
                     ListTables: col_1_values_results,
                     DeleteRows: id_count_results,
                     DropTable: simple_results,
    }

    assert q_type in parser_funcs, 'Unknown query type.'

    return parser_funcs[q_type](cur)


class Executer(object):
    def __init__(self, username, password, db, host, port=3306):
        self.conn = MySQLdb.connect(host=host, user=username, passwd=password,
                                    db=db, port=port)
        self.conn.autocommit = False
        self.cur = self.conn.cursor()

    def __call__(self, query):
        q_type = type(query)
        assert issubclass(q_type, Query), 'The passed parameter is not a query.'

        try:
            self.cur.execute(query.to_sql())

            if q_type in requires_commit:
                self.conn.commit()

            return format_result(q_type, self.cur)
        except Exception as err:
            # Note: This function may return a formated result, or it may reraise the error
            return handle_error(q_type, err)

    def many(self, queries):
        # TODO: Test the speed on this
        # TODO: Error handling
        # TODO: Test atomicity
        for q in queries:
            self.cur.execute(q.to_sql())
        self.conn.commit()

        return SQLExecutionResult(True, None)

    def raw(self, q_str):
        self.cur.execute(q_str)
        self.conn.commit()

        return SQLExecutionResult(True, None)

    def kill(self):
        self.cur.close()
        self.conn.close()


def run_tests(deps_provider):
    '''
    >>> create_table_query = CreateTable(
    ...      'test_users',
    ...      AutoIncrementColumn('id'),
    ...      [ ColumnDefinition('username', SQLType('VARCHAR', 30), True),
    ...        ColumnDefinition('drivers_licence_number', SQLType('VARCHAR', 30), True),
    ...        ColumnDefinition('first_name', SQLType('VARCHAR', 30), False),
    ...        ColumnDefinition('balance', SQLType('BIGINT'), False),
    ... ], if_not_exists=True )

    >>> res = ex(create_table_query)
    >>> res.success
    True
    >>> res.data is None
    True
    >>> res = ex(DescribeTable('test_users'))
    >>> res.success
    True
    >>> res.data
    {'Field': 'id', 'Type': 'bigint(20) unsigned', 'Null': 'NO', 'Key': 'PRI', 'Default': None, 'Extra': 'auto_increment'}
    {'Field': 'username', 'Type': 'varchar(30)', 'Null': 'YES', 'Key': 'UNI', 'Default': None, 'Extra': ''}
    {'Field': 'drivers_licence_number', 'Type': 'varchar(30)', 'Null': 'YES', 'Key': 'UNI', 'Default': None, 'Extra': ''}
    {'Field': 'first_name', 'Type': 'varchar(30)', 'Null': 'YES', 'Key': '', 'Default': None, 'Extra': ''}
    {'Field': 'balance', 'Type': 'bigint(20)', 'Null': 'YES', 'Key': '', 'Default': None, 'Extra': ''}

    >>> res = ex(AddTableColumn('test_users', ColumnDefinition('balance2', SQLType('BIGINT'), False)))
    >>> res.success
    True
    >>> res.data == None
    True
    >>> ex(DescribeTable('test_users')).data
    {'Field': 'id', 'Type': 'bigint(20) unsigned', 'Null': 'NO', 'Key': 'PRI', 'Default': None, 'Extra': 'auto_increment'}
    {'Field': 'username', 'Type': 'varchar(30)', 'Null': 'YES', 'Key': 'UNI', 'Default': None, 'Extra': ''}
    {'Field': 'drivers_licence_number', 'Type': 'varchar(30)', 'Null': 'YES', 'Key': 'UNI', 'Default': None, 'Extra': ''}
    {'Field': 'first_name', 'Type': 'varchar(30)', 'Null': 'YES', 'Key': '', 'Default': None, 'Extra': ''}
    {'Field': 'balance', 'Type': 'bigint(20)', 'Null': 'YES', 'Key': '', 'Default': None, 'Extra': ''}
    {'Field': 'balance2', 'Type': 'bigint(20)', 'Null': 'YES', 'Key': '', 'Default': None, 'Extra': ''}

    >>> res = ex(DropTableColumn('test_users', 'balance2'))
    >>> res.data == None
    True
    >>> ex(DropTableColumn('test_users', 'balance2')).success
    False

    >>> res = ex(InsertRows('test_users', ['username', 'first_name', 'balance'],
    ... [['tester', 'test', 500],
    ... ['tester2', 'two', 200],
    ... ]))
    >>> res.success and res.data == {'last_row_id': 1, 'row_count': 2}
    True

    >>> res = ex(UpdateRows('test_users',
    ...  QueryCriterion('eq', 'username', 'tester'),
    ...  {'balance': 0}
    ... ))
    >>> res.success and res.data == {'last_row_id': 0, 'row_count': 1}
    True


    >>> res = ex(SelectRows('test_users', ['username', 'balance'],
    ...   QueryCriterion('gt', 'balance', 10),
    ...   None,
    ...   None,
    ... ))
    >>> res.success
    True
    >>> res.data
    {'username': 'tester2', 'balance': 200}

    >>> res = ex(CountUniqueRows('test_users', 'balance', None))
    >>> res.success
    True
    >>> res.data
    {'balance': 0, '_count': 1}
    {'balance': 200, '_count': 1}

    >>> ex(CountRows('test_users', None)).data
    2

    >>> ex(CreateTable('test_users2', AutoIncrementColumn('id'), [], if_not_exists=True)).data == None
    True

    >>> ex(ListTables(None)).data
    ['test_users', 'test_users2']

    >>> ex(DeleteRows('test_users', QueryCriterion('eq', 'username', 'tester'))).data
    {'last_row_id': 0, 'row_count': 1}

    >>> res = ex.many([
    ...     CreateTable(
    ...           'test_users3',
    ...           AutoIncrementColumn('id'),
    ...           [ ColumnDefinition('username', SQLType('VARCHAR', 30), True),
    ...             ColumnDefinition('drivers_licence_number', SQLType('VARCHAR', 30), True),
    ...             ColumnDefinition('first_name', SQLType('VARCHAR', 30), False),
    ...             ColumnDefinition('balance', SQLType('BIGINT'), False),
    ...     ], if_not_exists=False),
    ...     InsertRows('test_users3', ['username', 'first_name', 'balance'],
    ...     [['tester1', 'test', 500],
    ...      ['tester2', 'two', 200],
    ...      ['tester3', 'two', 200],
    ...      ['tester4', 'two', 200],
    ...      ['tester5', 'two', 200],
    ...      ['tester6', 'two', 200],
    ...     ]),
    ... ])
    >>> res.data
    >>> res.success
    True
    '''

    import doctest, sys
    from seneca.engine.storage.mysql_executer import Executer
    ex = deps_provider(Executer)

    return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})



"""
This module contains types that describe Redis commands, but the commands don't
have an implementation, e.g. Get doesn't have a method like:
> Set('foo', 'bar').to_redis_string()
SET foo bar

This is important as these commands get run in completely different ways if they
are run against local memory, a remote Redis instance, or the transactional
backend.

Furthermore, these commands are not made to be subtyped for specific
implementations. This has been done for two reasons:

* Transactional backend is the primary use of this module, that backend wraps
local and redis backends, and commands may be attempted locally and ultimately
run remotely. In that scenario, a command would have to be awkwardly upcast to
a more general type, then downcast to the Redis backend subtype.

* All commands have a safe_run() method. This function runs one or more Redis
commands along with Python logic. The purpose of safe_run is to ensure that in
the context of a transaction, if a command is approved for a local write, it
won't fail when the transaction is committed and the command runs again against
the Redis backend. In these methods the command would have to pick the correct
implemenation for all other commands it runs based on its subtype. This would
also be awkward.

* NOTE: As a convention, command objects take regular Python types in their
constructors (not wrapped/RTypes like RScalarInt). The executers that run them
will generate RTypes for writing to db, and for returning local store and for
return values.

TODO: Switch from key input type to strings that more closely match RESP
"""
from abc import ABCMeta, abstractmethod
from typing import Union, List, Dict

from seneca.engine.util import auto_set_fields
from seneca.engine.storage.redisnap.resp_types import *

# TODO: Enforce run return types for run method in executer libs.


class Command(ReprIsConstructor):
    def run(self, ex):
        '''
        Execute the command. Passing an executer like this allows the command
        objects to exist completely separate from an executer. We'll implement
        at least two executers, direct Redis and nested snapshots, a.k.a.
        transactions with savepoints.

        It may seem oddly inside-out: obj.run(ex) -> ex(obj). It's done like
        this to enforce correct return type from executers that handle these
        commands.
        '''
        return ex(self)

    @abstractmethod
    def safe_run(self, local_ex):
        pass


# Inheriting these classes designates the subtype as performing these general
# types of operation on Redis.
class TypeCheck(Command): pass
class Read(Command): pass

class Write(Command):
    # def to_data_dependency(self):
    #     return None
    pass

class Mutate(Command):
    """
    Mutates alter data like writes, but unlike writes, if the mutated value
    is read, it will create a read dependancy on previous contracts, this
    is not the case with regular writes.
    """
    pass


#NOTE: This is a decorator function, not a class!
class run_methods_return_type:
    """
    TODO: This thing totally doesn't work. The annotation affect all classes in
    the hierarchy. Fix it.

    Ideally it should decorate both run()( and safe_run()
    """
    @auto_set_fields
    def __init__(self, t):
        pass

    def __call__(self, cls):
        cls.run.__annotations__['return'] = self.t
        return cls


def run_method_is_safe(cls):
    """
    This class decorator function designates that the the class it's run on
    has a safe run method i.e. it will never fail due to uncertain data on a
    Redis server. It actually sets the safe_run() method to run the code in the
    run() method.
    """
    # TODO: abstract this functionality and move to util
    if hasattr(cls, '__abstractmethods__'):
        abstr_methods = set(cls.__abstractmethods__)
        abstr_methods.remove('safe_run')
        cls.__abstractmethods__ = frozenset(abstr_methods)
    setattr(cls, 'safe_run', cls.run)
    return cls

################
# Key Commands #
################
@run_methods_return_type(type)
@run_method_is_safe
class Type(Command):
    @auto_set_fields
    def __init__(self, key):
        pass


@run_methods_return_type(bool)
@run_method_is_safe
class Exists(Command):
    @auto_set_fields
    def __init__(self, key):
        pass

    # to_data_dependency(self):
    #     return ExistentialDependancy(self.key)

@run_methods_return_type(type(None))
@run_method_is_safe
class Del(Write):
    @auto_set_fields
    def __init__(self, key):
        pass

    # to_data_dependency(self):
    #     return None

'''
Not yet implmented:
DUMP, EXPIRE, EXPIREAT, MIGRATE, MOVE, OBJECT, PERSIST, PEXPIRE, PEXPIREAT,
PTTL, RANDOMKEY, RENAME, RENAMENX, RESTORE, SORT, TOUCH, TTL, UNLINK, WAIT

Won't implement:
# KEYS not implemented, from docs Warning: consider KEYS as a command that should only be used in production environments with extreme care
# SCAN not implemented, will be hard to maintain the cursor and sequence of existing and newly added keys
'''


# TODO: must add constraints on inputs, annotations + typeguard @typechecked should be sufficient
###################
# String Commands #
###################
@run_methods_return_type(bool)
@run_method_is_safe
class AssertType(TypeCheck):
    @auto_set_fields
    def __init__(self, key: str, r_type: RESPType):
        pass


@run_methods_return_type(type(None))
class AppendNR(Mutate):
    @auto_set_fields
    def __init__(self, key: str, value: str):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RScalar))
        return ex(self)


@run_methods_return_type(bytes)
class Get(Read):
    @auto_set_fields
    def __init__(self, key: str):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RScalar))
        return ex(self)


@run_methods_return_type(type(None))
@run_method_is_safe
class SetNR(Write):
    @auto_set_fields
    def __init__(self, key: str, value: Union[str, float, int]):
        pass

# Note: Front end must convert Incr, Decr, and DecrBy to IncrBy
@run_methods_return_type(int)
class IncrByNR(Mutate):
    @auto_set_fields
    def __init__(self, key, amount: int):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RScalarInt))
        return ex(self)


#################
# Hash Commands #
#################
@run_methods_return_type(bool)
@run_method_is_safe
class AssertFieldType(TypeCheck):
    @auto_set_fields
    def __init__(self, key: str, field: str, r_type: RESPType):
        self.safe_run = self.run


@run_methods_return_type(bytes)
class HGet(Read):
    @auto_set_fields
    def __init__(self, key: str, field: str):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RHash))
        return ex(self)

#@run_methods_return_type(type(None))
class HSetNR(Write):
    @auto_set_fields
    def __init__(self, key: str, field: str, value: Union[str, float, int]):
        pass

    def safe_run(self, ex):
        # Just assert that the key has
        assert ex(AssertType(self.key, RHash))
        return ex(self)

class HExists(Read):
    @auto_set_fields
    def __init__(self, key: str, field: str):
        pass

    def safe_run(self, ex):
        # Just assert that the key has
        assert ex(AssertType(self.key, RHash))
        return ex(self)

#################
# List Commands #
#################
class LLen(Read):

    @auto_set_fields
    def __init__(self, key: str):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RList))
        return ex(self)


class LIndex(Read):
    @auto_set_fields
    def __init__(self, key: str, index: int):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RList))
        return ex(self)


class LSet(Write):
    @auto_set_fields
    def __init__(self, key: str, index: int, value: Union[str, float, int]):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RList))
        list_len = ex(LLen(self.key))
        assert list_len > self.index
        return ex(self)


class LPushNR(Mutate):
    @auto_set_fields
    def __init__(self, key: str, value: List[Union[str, float, int]]):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RList))
        return ex(self)


class RPushNR(Mutate):
    @auto_set_fields
    def __init__(self, key: str, value: List[Union[str, float, int]]):
        pass

    def safe_run(self, ex):
        raise NotImplementedError()
        assert ex(AssertType(self.key, RList))
        return ex(self)


class LPop(Mutate, Read):
    @auto_set_fields
    def __init__(self, key: str):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RList))
        return ex(self)


class RPop(Mutate, Read):
    @auto_set_fields
    def __init__(self, key: str):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RList))
        return ex(self)

class ZAddNR(Mutate):
    @auto_set_fields
    def __init__(self, key: str, members_and_scores: Dict[str, int]):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RSet))
        return ex(self)


class ZRemNR(Mutate):
    @auto_set_fields
    def __init__(self, key: str, members: List[str]):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RSet))
        return ex(self)


class ZRevRangeByScore(Read):
    @auto_set_fields
    def __init__(self, key: str, max: int, min: int, inclusive=(True, True), with_scores=False):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RSet))
        return ex(self)


class ZScore(Read):
    @auto_set_fields
    def __init__(self, key: str, member: str):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RSet))
        return ex(self)


class ZIncrByNR(Mutate):
    @auto_set_fields
    def __init__(self, key: str, amount:int, member: str):
        pass

    def safe_run(self, ex):
        assert ex(AssertType(self.key, RSet))
        return ex(self)


# TODO: command merging
'''
TODO:
* Bitmaps Commands
* hyperloglogs Commands
'''

"""
Note: This module currently holds shared tests for RediSnap
"""

def run_tests(deps_provider):
    import doctest, sys
    from seneca.engine.util import return_exception_tuple

    return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})



"""

Interface:

Main process
* new_session() (singleton):
  Returns a session object
  * Methods:
    * finalize()
    * new_transaction(rank_info):
      * creates a process to run executer, forks main process with latest list of other transactions
        * ExecuterBase is subscribed to latest transaction publisher
        * Will throw an exception immediately if a new block is published that causes a conflict.
        * will publish it's own transaction when it's soft-committed and close
      *



Executers, two sides talk over zmq socket:
  * ExClient, started in the client application
    * a connection to the ExServer to run queries
    * a connection to the ExServer that receives errors when higher priority transactions finish and break consistency of view
  * ExServer, forked from main process, it has:
    * a view of current data data, consisting of
      * a list of previously completed transactions, and reads from them
      * a connection to Redis read requests that aren't fulfilled by any transactions fall through to Redis
    * a subscription to receive any higher priority transactions that have completed to either
      * safely insert into its view (not breaking consistency), or
      * send an error message if view consistency in broken





# TODO(s):
* Finish granular address conflict detection
* Operation compaction
* Devise a method to poll for view inconsistencies while smart contracts are running without waiting for db access.
"""



'''This module decorates (in the design pattern sense) the mysql_intermediate
module, adding namespacing.

To convert:
 {'AddTableColumn',
 'CountRows',
 'CountUniqueRows',
 'CreateTable',
 'DeleteRows',
 'DescribeTable',
 'DropTable',
 'DropTableColumn',
 'InsertRows',

 'ListTables', (prefix, not table_name
 'SelectRows',
 'UpdateRows'}


 Just prepend table_name for everything except ListTables
'''

#from seneca.engine.util import run_super_first, auto_set_fields, intercalate
import seneca.engine.storage.mysql_intermediate as m_i


def name_space_isql(prefix, isql_obj):
    # TODO: Implement
    pass
    # Determine what kind of type

    # Run correct conversion

    # Return obj


# In tests, verify coverage of modified module
# : {k for (k,v) in mi.__dict__.items() if type(v) == type and issubclass(v, m_i.Query)}
def run_tests(_):
    '''
    >>> name_space_isql('', None)
    '''
    # TODO: Add teests.
    import doctest, sys, ast
    return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})



import random, seneca
from os import getenv as env
from dotenv import load_dotenv

def load_env():
    load_dotenv(dotenv_path='{}/../docker/redis.env'.format(path), override=True)

def get_redis_port():
    if env('CIRCLECI'):
        return 6379
    return env('REDIS_PORT', 6379)

def get_redis_password():
    if env('CIRCLECI'):
        return ''
    return env('REDIS_PASSWORD', '')

path = seneca.__path__[0]
load_env()

MASTER_DB = 0
DB_OFFSET = 1
NUM_CACHES = 2
CODE_OBJ_MAX_CACHE = 64



import redis, ast, marshal, array, copy, inspect, types, uuid, copy, ujson as json, sys, time
from seneca.constants.whitelists import ALLOWED_AST_TYPES, ALLOWED_IMPORT_PATHS, SAFE_BUILTINS, SENECA_LIBRARY_PATH
from seneca.constants.config import get_redis_port, get_redis_password, MASTER_DB, DB_OFFSET, CODE_OBJ_MAX_CACHE
from functools import lru_cache
from seneca.libs.metering.tracer import Tracer
import seneca, os
from os.path import join

class ReadOnlyException(Exception):
    pass

class CompilationException(Exception):
    pass

class SenecaInterpreter:

    exports = {}
    imports = {}
    loaded = {}
    cache = {}
    _is_setup = False
    concurrent_mode = True

    @classmethod
    def setup(cls, concurrent_mode=True):
        if not cls._is_setup:
            cls.r = redis.StrictRedis(host='localhost', port=get_redis_port(), db=MASTER_DB, password=get_redis_password())
            cls._is_setup = True
            cls.setup_tracer()
        cls.concurrent_mode = concurrent_mode

    @classmethod
    def setup_tracer(cls):
        seneca_path = seneca.__path__[0]
        path = join(seneca_path, 'constants', 'cu_costs.const')
        os.environ['CU_COST_FNAME'] = path
        cls.tracer = Tracer()

    @classmethod
    def teardown(cls):
        cls._is_setup = False

    @classmethod
    def get_code_obj(cls, fullname):
        assert cls._is_setup, "Must be set up to get_code_obj!!!!"
        code_obj = cls.r.hget('contracts', fullname)
        assert code_obj, 'User module "{}" not found!'.format(fullname)
        return marshal.loads(code_obj)

    @classmethod
    def get_code_str(cls, fullname):
        meta = cls.get_contract_meta(fullname)
        assert meta.get('code_str'), 'Cannot find original code string for module "{}" not found!'.format(fullname)
        return meta['code_str']

    @classmethod
    def get_contract_meta(cls, fullname):
        byte_str = cls.r.hget('contracts_meta', fullname)
        assert byte_str, 'Contract "{}" does not exist.'.format(fullname)
        meta = json.loads(byte_str)
        return meta

    @classmethod
    def set_code(cls, fullname, code_obj, code_str, author, keep_original=False):
        pipe = cls.r.pipeline()
        pipe.hset('contracts', fullname, marshal.dumps(code_obj))
        pipe.hset('contracts_meta', fullname, json.dumps({
            'code_str': code_str,
            'author': author,
            'timestamp': time.time()
        }))
        pipe.execute()

    @classmethod
    def remove_code(cls, fullname):
        pipe = cls.r.pipeline()
        pipe.hdel('contracts', fullname)
        pipe.hdel('contracts_meta', fullname)
        pipe.execute()

    @classmethod
    def code_obj_exists(cls, fullname) -> bool:
        return bool(cls.r.hget('contracts', fullname))

    @classmethod
    def assert_import_path(cls, import_path, module_name=None):
        if module_name == '*':
            raise ImportError('Not allowed to import *')
        elif module_name:
            import_path = '.'.join([import_path, module_name])
        if import_path.startswith(SENECA_LIBRARY_PATH):
            cls.exports[import_path] = True
            return True
        for path in ALLOWED_IMPORT_PATHS:
            if import_path.startswith(path):
                if len(import_path.split('.')) - len(path.split('.')) == 2:
                    if not cls.exports.get(import_path):
                        cls.imports[import_path] = True
                    return True
                else:
                    raise ImportError('Instead of importing the entire "{}" module, you must import each functions directly.'.format(import_path))
        raise ImportError('Cannot find module "{}" in allowed protected_imports'.format(import_path))

    @classmethod
    def validate(cls):
        for import_path in cls.imports:
            if not cls.exports.get(import_path):
                raise CompilationException('Forbidden to import "{}"'.format(
                    import_path))

    @classmethod
    def parse_ast(cls, code_str, protected_variables=[]):

        tree = ast.parse(code_str)
        protected_variables += ['export']
        current_ast_types = set()
        prevalidated = copy.deepcopy(tree)
        prevalidated.body = []

        for idx, item in enumerate(ast.walk(tree)):

            if isinstance(item, ast.Name):
                if cls.is_system_variable(item.id):
                    raise CompilationException('Not allowed to read "{}"'.format(item.id))

            # Restrict __attr__ to be accessed
            if isinstance(item, ast.Attribute):
                if cls.is_system_variable(item.attr):
                    raise CompilationException('Not allowed to read "{}"'.format(item.attr))

            # Restrict protected_imports to ones in ALLOWED_IMPORT_PATHS
            elif isinstance(item, ast.Import):
                module_name = item.names[0].name
                cls.assert_import_path(module_name)
                prevalidated.body.append(item)

            elif isinstance(item, ast.ImportFrom):
                module_name = item.names[0].name
                cls.assert_import_path(item.module, module_name=module_name)
                prevalidated.body.append(item)

            # Restrict variable assignment
            elif isinstance(item, ast.Assign):
                for target in item.targets:
                    cls.check_protected(target, protected_variables)

            elif isinstance(item, ast.AugAssign):
                cls.check_protected(item.target, protected_variables)

            elif isinstance(item, ast.FunctionDef):
                for fn_item in item.body:
                    if isinstance(fn_item, (ast.Import, ast.ImportFrom)):
                        raise ImportError('Cannot import modules inside a function!')

                prevalidated.body.append(item)
            current_ast_types.add(type(item))

        illegal_ast_nodes = current_ast_types - ALLOWED_AST_TYPES
        assert not illegal_ast_nodes, 'Illegal AST node(s) in module: {}'.format(
            ', '.join(map(str, illegal_ast_nodes)))

        return tree, prevalidated

    @staticmethod
    def is_system_variable(v):
        return v.startswith('__') and v.endswith('__')

    @classmethod
    def check_protected(cls, target, protected_variables):
        if isinstance(target, ast.Subscript):
            return
        if target.id in protected_variables \
            or target.id in [k.rsplit('.', 1)[-1] for k in cls.imports.keys()]:
            raise ReadOnlyException('Cannot assign value to "{}" as it is a read-only variable'.format(target.id))

    @classmethod
    def execute(cls, code, scope={}, is_main=True):
        scope.update({
            'export': Export(),
            '__builtins__': SAFE_BUILTINS,
            '__use_locals__': False,
        })
        if is_main:
            cls.loaded['__main__'] = scope
        exec(code, scope)
        if is_main:
            cls.validate()

    @classmethod
    @lru_cache(maxsize=CODE_OBJ_MAX_CACHE)
    def get_cached_code_obj(cls, module_path):
        cls.assert_import_path(module_path)
        module = module_path.rsplit('.', 1)
        code_str = '''
result = {}()
        '''.format(module[1])
        code_obj = compile(code_str, '__main__', 'exec')
        import_obj = compile('from {} import {}'.format(module[0], module[1]), '__main__', 'exec')
        return code_obj, import_obj

    @classmethod
    def execute_function(cls, module_path, author, sender, stamps, *args, **kwargs):
        scope = {
            'rt': { 'author': author, 'sender': sender, 'contract': module_path },
            '__builtins__': SAFE_BUILTINS,
            '__args__': args,
            '__kwargs__': kwargs,
            '__use_locals__': True
        }
        code_obj, import_obj = cls.get_cached_code_obj(module_path)
        cls.loaded['__main__'] = scope
        exec(import_obj, scope)
        cls.tracer.set_stamp(stamps)
        cls.tracer.start()
        exec(code_obj, scope)
        cls.tracer.stop()
        # print('xxx',cls.tracer.get_stamp_used())
        return {
            'status': 'success',
            'output': scope.get('result'),
            'remaining_stamps': stamps - cls.tracer.get_stamp_used()
        }

class ScopeParser:

    def set_scope(self, fn, args, kwargs):
        fn.__globals__.update(SenecaInterpreter.loaded['__main__'])
        fn.__globals__['rt']['contract'] = fn.__module__
        if fn.__globals__.get('__use_locals__'):
            if fn.__globals__.get('__args__'): args = fn.__globals__['__args__']
            if fn.__globals__.get('__kwargs__'): kwargs = fn.__globals__['__kwargs__']
        return args, kwargs

    def set_scope_during_compilation(self, fn):
        self.module = '.'.join([fn.__module__, fn.__name__])
        SenecaInterpreter.exports[self.module] = True

class Export(ScopeParser):
    def __call__(self, fn):
        if not fn.__module__: return
        self.set_scope_during_compilation(fn)
        def _fn(*args, **kwargs):
            args, kwargs = self.set_scope(fn, args, kwargs)
            return fn(*args, **kwargs)
        return _fn



import sys, importlib, warnings
from seneca.engine.module import SenecaFinder, RedisFinder
from seneca.engine.interpreter import SenecaInterpreter

class SenecaInterface(SenecaInterpreter):
    """
        High level API for interacting with Seneca Smart Contracts
    """

    def __init__(self, concurrent_mode=True):
        if not isinstance(sys.meta_path[2], RedisFinder):
            self.old_sys_path = sys.meta_path
            sys.meta_path = [sys.meta_path[2], SenecaFinder(), RedisFinder()]
        SenecaInterpreter.setup(concurrent_mode)

    def __enter__(self, *args, **kwargs):
        self.old_concurrent_mode = SenecaInterpreter.concurrent_mode
        return self

    def __exit__(self, type, value, traceback):
        SenecaInterpreter.concurrent_mode = self.old_concurrent_mode
        return False

    def compile_code(self, code_str, scope={}):
        tree, prevalidated = self.parse_ast(code_str, protected_variables=list(scope.keys()))
        prevalidated_obj = compile(prevalidated, filename='__main__', mode="exec")
        self.execute(prevalidated_obj, scope)
        code_obj = compile(tree, filename='__main__', mode="exec")
        return code_obj

    def execute_code_str(self, code_str, scope={'rt': {'sender': 'anonymous', 'author': 'anonymous', 'contract': 'arbitrary'}}):
        try:
            code_obj = self.compile_code(code_str, scope)
            return self.execute(code_obj, scope)
        except:
            SenecaInterpreter.imports = {}
            raise

    def publish_code_str(self, fullname, author, code_str, keep_original=False, scope={}):
        try:
            assert not self.r.hexists('contracts', fullname), 'Contract "{}" already exists!'.format(fullname)
            with SenecaInterface(False) as interface:
                code_obj = self.compile_code(code_str, scope={'rt': {'author': author, 'contract': fullname}})
                self.set_code(fullname, code_obj, code_str, author, keep_original)
        except:
            SenecaInterpreter.imports = {}
            raise



from setuptools import setup, find_packages
from setuptools.extension import Extension
import os

major = 0

def get_version_number():
    if os.getenv('CIRCLECI'):
        minor, patch = divmod(int(os.getenv('CIRCLE_BUILD_NUM')), 360)
        ver = '{}.{}.{}'.format(major, minor, patch)
        return ver
    else:
        return '{}.1.0'.format(major)

__version__ = get_version_number()

print('#' * 128)
print('\n    VERSION = {}\n'.format(__version__))
print('#' * 128)

requirements = [
    'redis==2.10.6',
    'python-dotenv==0.9.1',
    'ujson==1.35'
]

setup(
    name='seneca',
    version=__version__,
    description='Python-based smart contract language and interpreter.',
    entry_points={
        'console_scripts': ['seneca=seneca.cli:main'],
    },
    packages=find_packages(),
    install_requires=requirements,
    url='https://github.com/Lamden/seneca',
    author='Lamden',
    author_email='team@lamden.io',
    classifiers=[
        'Programming Language :: Python :: 3.6',
    ],
    zip_safe=False,
    include_package_data=True,
    ext_modules=[
        Extension('seneca.libs.metering.tracer', sources = ['seneca/libs/metering/tracer.c'])
    ]
)



from unittest import TestCase
from seneca.engine.interface import SenecaInterface
from seneca.engine.interpreter import SenecaInterpreter, ReadOnlyException, CompilationException
from os.path import join
from tests.utils import captured_output, TestInterface
import redis, unittest, seneca

test_contracts_path = seneca.__path__[0] + '/../test_contracts/'

class TestScope(TestInterface):

    def test_scope(self):
        """
            Importing exported functions should pass
        """
        self.si.execute_code_str("""
from test_contracts.good import one_you_can_export, one_you_can_also_export, one_you_can_also_also_export
one_you_can_export()
one_you_can_also_export()
one_you_can_also_also_export()
        """)

    def test_scope_fail(self):
        """
            Importing protected functions should fail
        """
        with self.assertRaises(CompilationException) as context:
            self.si.execute_code_str("""
from test_contracts.good import one_you_cannot_export
            """)

    def test_globals(self):
        with captured_output() as (out, err):
            self.si.execute_code_str("""
from test_contracts.reasonable import reasonable_call
print(reasonable_call())
            """, {'rt': {'sender':'123'}})
            self.assertEqual(out.getvalue().strip(), 'sender: 123, contract: test_contracts.reasonable')

    def test_globals_redis(self):
        bk_info = {'sbb_idx': 2, 'contract_idx': 12}
        rt_info = {'rt': {'sender': 'davis', 'author': 'davis'}}
        all_info = {**bk_info, **rt_info}
        with open(join(test_contracts_path, 'sample.sen.py')) as f:
            self.si.publish_code_str('sample', 'davis', f.read(), keep_original=True)
        with captured_output() as (out, err):
            self.si.execute_code_str("""
from seneca.contracts.sample import do_that_thing
print(do_that_thing())
            """, all_info)
            self.assertEqual(out.getvalue().strip(), 'sender: davis, author: davis')

    def test_execute_function(self):
        result = self.si.execute_function('test_contracts.reasonable.call_with_args',
            'me', 'also_me', 10000, 'it is required', not_required='it is not requried')
        self.assertEqual(result['status'], 'success')

    def test_execute_function_invalid(self):
        with self.assertRaises(ImportError) as context:
            result = self.si.execute_function('seneca.engine.util.make_n_tup',
                'me', 'also_me', 10000, {'x': 'y'})
            print('Should not print this: ', result)

    def test_execute_function_out_of_gas(self):
        with self.assertRaises(AssertionError) as context:
            result = self.si.execute_function('test_contracts.reasonable.call_with_args',
                'me', 'also_me', 5, 'it is required', not_required='it is not requried')


if __name__ == '__main__':
    unittest.main()



'''
# Simple tabular datastore #
* Uses myrocks to support SQL operations on rocksdb
* Right now relying on yet to be built execution environment to force squential evaluation of contracts
  * Not doing any concurrency control/transactions. Consistency maintained by that sequential evaluation

* Tabular2
  * When delete column, delete from table object
  * have to change the way we auto increment id, has to be compatible with rollback
  * Works with both mysql-base and mysql-pit

*TODOs
  * upsert, i.e. ON DUPLICATE KEY UPDATE
  * list tables
  * Type annotations
'''

from itertools import zip_longest
import warnings
import datetime
from datetime import datetime, timedelta

from seneca.engine.util import auto_set_fields, add_methods, add_method_as, filter_split, assert_len
import seneca.engine.storage.mysql_intermediate as isql
from seneca.engine.storage.mysql_base import FixedStr, cast_py_to_sql
#from inflection import  underscore, camelize

# Convenience functions for dealing with lists, dicts, and objects
def m_subscript(sub_able, sub):
    '''Maybe subscript, if a subscript exists at the specified index or key, it
    will be returned, else, None will be returned, errors are suppressed.'''
    try:
        return sub_able[sub]
    except (IndexError, KeyError) as e:
        return None


def m_getattr(sub_able, sub):
    '''Maybe getattr, if an attr exists it will be returned, else, None will be
    returned, errors are suppressed.'''
    try:
        return getattr(sub_able, sub)
    except AttributeError as e:
        return None


def get_exactly_one(xs):
    assert_len(1, xs)
    return xs[0]


def filter_type(xs, t):
    return [x for x in xs if type(x) == t]


def table_name_from_cs(cs):
    return require_from_cs(cs, Table, 'sql_name')()


def optional_from_cs(stack, type_, attr):
    return m_getattr(m_subscript(filter_type(stack, type_), 0), attr)


def require_from_cs(cs, t, attr):
    return getattr(get_exactly_one(filter_type(cs, t)), attr)


def str_len(l):
    '''Type constructor function, made to look like builtin str()'''
    return FixedStr.Len(l)


### Shared methods, added to classes via decorator, treated like single-method traits ###
def where(self, where_criterion):
    return WhereClause(self.call_stack.copy(), where_criterion)


def order_by(self, order_by_column_name, desc=False):
    return OrderBy(self.call_stack.copy(), order_by_column_name, desc=desc)


def limit(self, count_limit):
    return LimitTo(self.call_stack.copy(), count_limit)


def terminal_where(self, where_criterion):
    # A where clause without order_by or limit, effective terminates choices in call chain (only .run)
    return TerminalWhereClause(self.call_stack.copy(), where_criterion)
# End of single-method traits


def execute_sql_query(executer, isql_obj):
    res = executer(isql_obj)

    if not res.success:
        raise Exception(res.data)
    else:
        return res.data


class Column(object):
    @auto_set_fields
    def __init__(self, name, type=None, unique=False):
        pass

    # Custom methods make the results of comparisons in query components
    # e.g (Column('some_column') != None).to_sql() -> 'some_column IS NOT NULL'
    def __eq__(self, other):
        return isql.QueryCriterion('eq', self.name, other)

    def __ne__(self, other):
        return isql.QueryCriterion('ne', self.name, other)

    def __lt__(self, other):
        return isql.QueryCriterion('lt', self.name, other)

    def __gt__(self, other):
        return isql.QueryCriterion('gt', self.name, other)

    def __le__(self, other):
        return isql.QueryCriterion('le', self.name, other)

    def __ge__(self, other):
        return isql.QueryCriterion('ge', self.name, other)

    def to_intermediate_def(self):
        # TODO: Change this into a proper raise of custom exception class
        assert self.type, "When defining table column, a type must be declared"
        sql_type = isql.SQLType.from_python_type(self.type)

        return isql.ColumnDefinition(self.name, sql_type, unique=self.unique)

    @classmethod
    def from_sql_describe_row(cls, row):
        '''Alternate constructor: Instead of constructing from user provided
        parameters, it runs a DESCRIBE <TABLE>; query, and construct Column from
        the data it gets back.'''

        # XXX: This may break if sql is updated or maria vs mysql, or something
        name = row['Field']

        if row['Type'].lower() == 'bigint(20) unsigned' and \
                     row['Extra'].lower() == 'auto_increment':
            return AutoIncrementColumn(name)
        else:
            type_ = isql.SQLType.from_db_describe_str(row['Type'])
            unique = row['Key'].lower() == 'uni'
            return cls(name, type=type_, unique=unique)


class AutoIncrementColumn(Column):
    '''Special instance of Column with automatically incrementing bigint,
    intended for use as primary keys.'''
    @auto_set_fields
    def __init__(self, name):
        pass

    def to_intermediate_def(self):
        return isql.AutoIncrementColumn(self.name)


def and_(*args):
    '''Function constructs compound where clauses, matches SQLAlchemy API.'''
    return isql.AndedCriteria(args)

def not_(c):
    '''Function negates where clauses, matches SQLAlchemy API.'''
    return isql.InvertedCriterion(c)

def or_(*args):
    '''Function constructs compound where clauses, matches SQLAlchemy API.'''
    return isql.OredCriteria(args)


# NOTE: The following classes are used to build sql-like method chains. They
# each represent a destinct part of an SQL query. When chained, they follow the
# same rules as real SQL regarding syntax and order of operations e.g.
# 'order by' cannot come before 'where'.
# Different parts of SQL must be in order and are often optional (i.e.
# not required), e.g. 'where ...', 'order by', 'limit'. Generally speaking,
# these classes each have an ever narrowing set of available methods, e.g.
# select may be followed by .where(), order_by(), limit(), or run(), but
# .order_by(...) may only be followed by .limit() and .run() methods available,
# per SQL syntax, .where() would have had to come before order_by(). In contrast
# .select() will return an object with all previously mentioned methods available.
class QueryComponent(object):
    '''Base class for sql method chains.'''
    def __init__(self, call_stack, *args, **kwargs):
        self.call_stack = call_stack
        self.call_stack.append(self)
        self._supplemental_init(*args, **kwargs)

    def _supplemental_init(self, *args, **kwargs):
        '''Many constructor for sql methods have special arguments and behaviors.
        Structuring init in 2 functions allows classes to inherit and use a common
        __init__ to handle the callstack, but still have unique supplemental
        behaviors layered on top.'''
        pass


    def to_isql(self):
        ''' This method starts the process of unwinding the method chain
        callstack, it finds the sql verb (e.g. select(), update(), insert(), or
        delete()), does some light validation, then uses the verb's
        to_intermediate to turn the entire callstack into a representation of a
        SQL query.'''
        tlqcs = [x for x in self.call_stack if issubclass(type(x), SQLVerb)]
        assert_len(1, tlqcs) # There should only ever be one method/verb per query.
        sql_verb = tlqcs[0]
        # The SQL method/verb is the only query component type that can generate
        # a representation of the query
        return sql_verb.to_intermediate(self.call_stack)

    def run(self, ex):
        return execute_sql_query(ex, self.to_isql())

    def to_sql(self):
        return self.to_isql().to_sql()

    def __str__(self):
        d1 = self.__dict__.copy()
        d1.pop('call_stack')
        return '%s(%s)' % (type(self).__name__, str(d1))

    def __repr__(self):
        return self.__str__()


class LimitTo(QueryComponent):
    ''' Limit must be the last statement in our standard SQL queries, it
    inherits the .run() method from QueryComponent, and can't do anything else.
    '''
    @auto_set_fields
    def _supplemental_init(self, count_limit):
        # TODO: Validate count_limit value
        pass


class TerminalWhereClause(QueryComponent):
    '''Simplified where clause that doesn't have any methods except .run() so
    it effectively ends the chain of methods.'''
    @auto_set_fields
    def _supplemental_init(self, where_criterion):
        # TODO: Validate where_criterion value
        pass


@add_methods(limit)
class OrderBy(QueryComponent):
    @auto_set_fields
    def _supplemental_init(self, column_name, desc=False):
        pass


@add_methods(limit, order_by)
class WhereClause(QueryComponent):
    @auto_set_fields
    def _supplemental_init(self, where_criterion):
        #TODO: Verify not empty
        pass


class SQLVerb(QueryComponent):
    ''' Abstract class, todo, use ABC and @abstractmethod '''
    @auto_set_fields
    def _supplemental_init(self, *where_criteria):
        raise NotImplementedError("Class should never be instantiated")


# This verb has run() but no other chainable methods.
class AddTableColumn(SQLVerb):
    @auto_set_fields
    def _supplemental_init(self, column_def):
        pass

    def to_intermediate(self, full_call_stack):
        table_name = table_name_from_cs(full_call_stack)
        isql_column_def = self.column_def.to_intermediate_def()

        return isql.AddTableColumn(table_name, isql_column_def)


class DropTableColumn(SQLVerb):
    @auto_set_fields
    def _supplemental_init(self, column_name):
        pass

    def to_intermediate(self, full_call_stack):
        # TODO: Dedupe this section from other Classes. Ideally with __lt__ implemented for automatic sorting.
        # and individual to_intermediate methods for all classes so just sort and map that over list.
        table_name = table_name_from_cs(full_call_stack)
        column_name = require_from_cs(full_call_stack, DropTableColumn, 'column_name')

        return isql.DropTableColumn(table_name, column_name)


@add_method_as(terminal_where, 'where')
class CountRows(SQLVerb):
    # Overwrite supplemental with noop
    def _supplemental_init(self, *args):
        pass

    def to_intermediate(self, call_stack):
        # TODO: Dedupe this section from other Classes. Ideally with __lt__ implemented for automatic sorting.
        table_name = table_name_from_cs(call_stack)
        criterion = optional_from_cs(call_stack, TerminalWhereClause, 'where_criterion')
        return isql.CountRows(table_name, criterion)


@add_method_as(terminal_where, 'where')
class CountUniqueRows(SQLVerb):
    @auto_set_fields
    def _supplemental_init(self, column_name):
        pass

    def to_intermediate(self, call_stack):
        # TODO: Dedupe this section from other Classes. Ideally with __lt__ implemented for automatic sorting.
        # and individual to_intermediate methods for all classes so just sort and map that over list.
        table_name = table_name_from_cs(call_stack)
        unique_column_name = require_from_cs(call_stack, CountUniqueRows, 'column_name')
        criterion = optional_from_cs(call_stack, TerminalWhereClause, 'where_criterion')
        return isql.CountUniqueRows(table_name, unique_column_name, criterion)


@add_methods(where, order_by, limit)
class DeleteRows(SQLVerb):
    # Overwrite supplemental with noop
    def _supplemental_init(self, *args):
        pass

    def to_intermediate(self, full_call_stack):
        table_name = table_name_from_cs(full_call_stack)
        criterion = optional_from_cs(full_call_stack, WhereClause, 'where_criterion')
        order_by = optional_from_cs(full_call_stack, OrderBy, 'column_name')
        order_desc = optional_from_cs(full_call_stack, OrderBy, 'desc')
        limit = optional_from_cs(full_call_stack, LimitTo, 'count_limit')
        return isql.DeleteRows(table_name,
                               criterion,
                               order_by=order_by,
                               order_desc=order_desc,
                               limit=limit)

class InsertRows(SQLVerb):
    @auto_set_fields
    def _supplemental_init(self, list_column_val_dicts):
        pass

    @staticmethod
    def to_intermediate(full_call_stack):
        table_name = table_name_from_cs(full_call_stack)
        list_column_val_dicts = require_from_cs(full_call_stack, InsertRows, 'list_column_val_dicts')
        tab_kv = isql.TabularKVs.from_dicts(list_column_val_dicts)
        return isql.InsertRows(table_name, *tab_kv.to_klist_vlists())


@add_methods(where, order_by, limit)
class UpdateRows(SQLVerb):
    @auto_set_fields
    def _supplemental_init(self, column_val_dict):
        pass

    @staticmethod
    def to_intermediate(full_call_stack):
        table_name = table_name_from_cs(full_call_stack)
        column_value_dict = require_from_cs(full_call_stack, UpdateRows, 'column_val_dict')
        criterion = optional_from_cs(full_call_stack, WhereClause, 'where_criterion')
        order_by = optional_from_cs(full_call_stack, OrderBy, 'column_name')
        order_desc = optional_from_cs(full_call_stack, OrderBy, 'desc')
        limit = optional_from_cs(full_call_stack, LimitTo, 'count_limit')

        return isql.UpdateRows(table_name,
                               criterion,
                               column_value_dict,
                               order_by=order_by,
                               order_desc=order_desc,
                               limit=limit)

class SetRows(SQLVerb):
    @auto_set_fields
    def _supplemental_init(self, column_val_list):
        pass

    @staticmethod
    def to_intermediate(full_call_stack):
        table_name = table_name_from_cs(full_call_stack)
        column_val_list = require_from_cs(full_call_stack, SetRows, 'column_val_list')
        return isql.SetRows(table_name, column_val_list)


@add_methods(where, order_by, limit)
class SelectRows(SQLVerb):
    def _supplemental_init(self, *field_names):
        self.field_names = list(field_names)

    @staticmethod
    def to_intermediate(full_call_stack):
        table_name = table_name_from_cs(full_call_stack)
        column_names = optional_from_cs(full_call_stack, SelectRows, 'field_names')
        criterion = optional_from_cs(full_call_stack, WhereClause, 'where_criterion')
        order_by = optional_from_cs(full_call_stack, OrderBy, 'column_name')
        order_desc = optional_from_cs(full_call_stack, OrderBy, 'desc')
        limit = optional_from_cs(full_call_stack, LimitTo, 'count_limit')
        return isql.SelectRows(table_name,
                               column_names,
                               criterion,
                               order_by=order_by,
                               order_desc=order_desc,
                               limit=limit)


class Table(object):
    # TODO: create conveniently with id field already specced
    def __init__(self, name, primary_column, other_columns):
        self.call_stack = [self]
        self._name = name
        self.primary_key_column = primary_column
        self.other_columns = other_columns

        native_fields = set(dir(self))
        if primary_column:
            all_columns = [primary_column] + other_columns
        else:
            all_columns = other_columns
        column_names = set(map(lambda x: x.name, all_columns))

        namespace_conflicts = native_fields & column_names
        assert not namespace_conflicts, 'A forbidden column name has been used.'

        for c in all_columns:
            setattr(self, c.name, c)

    @classmethod
    def from_existing(cls, name):
        def run(ex):
            # TODO: return runable
            column_descriptions = execute_sql_query(ex, isql.DescribeTable(name))

            # Find primary key column and separate from others
            pkey_col_dict_list, other_cols_dict_list = \
              filter_split(lambda r: r['Key'].upper() == 'PRI', column_descriptions)
            if len(pkey_col_dict_list) == 0:
                pkey_col = None
            elif len(pkey_col_dict_list) == 1:
                pkey_col = Column.from_sql_describe_row(pkey_col_dict_list[0])
            else:
                raise Exception("Error, there should only every be one primary key column description")

            return cls(name,
                       pkey_col,
                       [Column.from_sql_describe_row(x) for x in other_cols_dict_list]
                       )

        return SimpleRunnable(run)


    def sql_name(self):
        #TODO: Implement this
        return self._name

    def insert(self, data):
        return InsertRows(self.call_stack.copy(), data)

    def select(self, *field_names):
        return SelectRows(self.call_stack.copy(), *field_names)

    def update(self, column_val_dict):
        return UpdateRows(self.call_stack.copy(), column_val_dict)

    def set(self, column_val_dict):
        return SetRows(self.call_stack.copy(), column_val_dict)

    def get(self, field_name):
        return self.select('v', 't').where(self.k == field_name)

    def delete(self):
        return DeleteRows(self.call_stack.copy())

    def count(self):
        return CountRows(self.call_stack.copy())

    def count_unique(self, column_name):
        return CountUniqueRows(self.call_stack.copy(), column_name)

    def add_column(self, *args, **kwargs):
        column_def = Column(*args, **kwargs)
        return AddTableColumn(self.call_stack.copy(), column_def)

    def drop_column(self, column_name):
        return DropTableColumn(self.call_stack.copy(), column_name)

    def __str__(self):
        d1 = self.__dict__.copy()
        d1.pop('call_stack')
        return '%s(%s)' % (type(self).__name__, str(d1))

    def __repr__(self):
        return self.__str__()

    def _to_intermediate_create(self, if_not_exists):
        return isql.CreateTable(self._name,
                                self.primary_key_column.to_intermediate_def(),
                                list(map(
                                  lambda x: x.to_intermediate_def(),
                                  self.other_columns
                                )),
                                if_not_exists=if_not_exists
                               )

    def create_table(self, if_not_exists=False):
        return RunnableISQL(self._to_intermediate_create(if_not_exists))

    def _to_intermediate_drop(self):
        return isql.DropTable(self._name)

    def drop_table(self):
        # TODO: if exists
        return RunnableISQL(self._to_intermediate_drop())


class SimpleRunnable(object):
    ''' Bare bones class with run method.
    '''
    @auto_set_fields
    def __init__(self, run_func):
        pass

    def run(self, *args, **kwargs):
        return self.run_func(*args, **kwargs)


class RunnableISQL(object):
    @auto_set_fields
    def __init__(self, isql, results_parser=lambda x:x):
        pass

    def to_isql(self):
        return self.isql

    def run(self, ex):
        return self.results_parser(execute_sql_query(ex, self.isql))

    def to_sql(self):
        return self.isql.to_sql()


def run_tests(deps_provider):
    '''
    TODO: Tests to validate interface of return values from update, select, etc.
    TODO: Decide whether or not object properties can be passed to select like strings

    >>> u = Table('users', AutoIncrementColumn('id'),[
    ...    Column('first_name', str),
    ...    Column('last_name', str),
    ...    Column('balance', int),
    ...    Column('creation_date', datetime)
    ... ])

    >>> u.create_table(if_not_exists=False).run(ex)
    >>> x = u.select().where(u.first_name != None).run(ex)
    >>> type(x) == TabularKVs
    True
    >>> len(x) == 0
    True

    >>> _ = u.insert([
    ...  {'first_name': 'Test', 'last_name': 'User','balance': 10},
    ...  {'first_name': 'Test2', 'last_name': 'User','balance': 20},
    ...  {'first_name': 'Test3', 'last_name': 'User','balance': 30},
    ... ]).run(ex)

    >>> print(repr(u.select().run(ex)))
    {'id': 1, 'first_name': 'Test', 'last_name': 'User', 'balance': 10, 'creation_date': None}
    {'id': 2, 'first_name': 'Test2', 'last_name': 'User', 'balance': 20, 'creation_date': None}
    {'id': 3, 'first_name': 'Test3', 'last_name': 'User', 'balance': 30, 'creation_date': None}


    # Test ordering of select
    >>> u.select('id', 'first_name').order_by('id', desc=True).run(ex)
    {'id': 3, 'first_name': 'Test3'}
    {'id': 2, 'first_name': 'Test2'}
    {'id': 1, 'first_name': 'Test'}

    # Test counts
    >>> u.count().run(ex)
    3
    >>> u.count().where(u.first_name == 'Test2').run(ex)
    1
    >>> u.count().where(u.first_name != 'Test2').run(ex)
    2

    # Test count uniques
    >>> u.count_unique('first_name').run(ex)
    {'first_name': 'Test', '_count': 1}
    {'first_name': 'Test2', '_count': 1}
    {'first_name': 'Test3', '_count': 1}


    # Test full update of all rows
    >>> _ = u.update({'balance': 1000}).run(ex)
    >>> print(repr(u.select('balance').run(ex)))
    {'balance': 1000}
    {'balance': 1000}
    {'balance': 1000}


    # Modify columns
    >>> u.add_column('nick_name', str).run(ex)
    >>> u.add_column('unique_nick_name', str_len(30), True).run(ex)
    >>> u.select('nick_name', 'unique_nick_name').run(ex)
    {'nick_name': None, 'unique_nick_name': None}
    {'nick_name': None, 'unique_nick_name': None}
    {'nick_name': None, 'unique_nick_name': None}

    # TODO: test to make sure unique is enforced

    # Test a delete, where clause, and an or_
    >>> _ = u.delete().where(
    ...    or_(
    ...        u.first_name == 'Test',
    ...        Column('first_name') == 'Test3'
    ...    )
    ... ).run(ex)
    >>> u.select('first_name').run(ex)
    {'first_name': 'Test2'}

    # Test pull existing table by name
    >>> t2 = Table.from_existing('users').run(ex)
    >>> t2.select('first_name').run(ex)
    {'first_name': 'Test2'}
    '''
    import doctest, sys
    from seneca.engine.storage.mysql_executer import Executer
    from seneca.engine.storage.mysql_base import TabularKVs

    ex = deps_provider(Executer)

    return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})



"""
"""
import redis
from functools import wraps
from seneca.engine.util import grouper

from seneca.engine.storage.redisnap.commands import *
import seneca.engine.storage.redisnap.resp_types as rtype
#from seneca.engine.storage.redisnap.addresses import *
from seneca.engine.storage.redisnap.backend_abc import ExecuterBase as abc_executer

def bytes_to_rscalar(b):
    if b:
        b = b.decode("utf-8")
    return rtype.make_rscalar(b)


class Executer(abc_executer):
    '''
    Maps command objects to actual Redis commands and runs them, leans heavily
    on redis.py

    TODO: We should efficiently track collisions and decide whether we want to
    use a log of transactions to commit, or create ops from the stored data
    '''
    def __init__(self, host, port):
        self._redis_executer = redis.StrictRedis(host=host, port=port)

    # TODO: decide if this is worth implemeneting
    def purge(self):
       self._redis_executer.flushdb()

    def exists(self, cmd):
        return self._redis_executer.exists(cmd.key)

    def type(self, cmd):
        return rtype.from_resp_str(self._redis_executer.type(cmd.key).decode("utf-8"))

    def asserttype(self, cmd):
        return isinstance(self.get(cmd), cmd.r_type)

    def get(self, cmd):
        return bytes_to_rscalar(self._redis_executer.get(cmd.key))


    def setnr(self, cmd):
        self._redis_executer.set(cmd.key, cmd.value)

    def incrbynr(self, cmd):
        try:
            self._redis_executer.incr(cmd.key, cmd.amount)
        except redis.exceptions.ResponseError as e:
            if str(e) == 'value is not an integer or out of range':
                raise RedisVauleTypeError('Existing value has wrong type.')
            else:
                raise

    def appendnr(self, cmd):
        self._redis_executer.append(cmd.key, cmd.value)


    def hget(self, cmd):
        return bytes_to_rscalar(self._redis_executer.hget(cmd.key, cmd.field))

    def hsetnr(self, cmd):
        self._redis_executer.hset(cmd.key, cmd.field, cmd.value)

    def hexists(self, cmd):
        return self._redis_executer.hexists(cmd.key, cmd.field)

    def del_(self, cmd):
        self._redis_executer.delete(cmd.key)


    def lindex(self, cmd):
        return bytes_to_rscalar(self._redis_executer.lindex(cmd.key, cmd.index))


    def lset(self, cmd):
        try:
            self._redis_executer.lset(cmd.key, cmd.index, cmd.value)
        except redis.exceptions.ResponseError as e:
            if e.args[0] == 'no such key':
                raise RedisKeyTypeError('Cannot LSet an nonexistent key.')
            elif e.args[0] == 'index out of range':
                return RedisListOutOfRange('Index out of range.')
            else:
                raise


    def lpushnr(self, cmd):
        self._redis_executer.lpush(cmd.key, *cmd.value)


    def rpushnr(self, cmd):
        self._redis_executer.rpush(cmd.key, *cmd.value)


    def _pop_base(self, method_name, cmd):
        return bytes_to_rscalar(getattr(self._redis_executer, method_name)(cmd.key))

    def lpop(self, cmd):
        return self._pop_base('lpop', cmd)

    def rpop(self, cmd):
        return self._pop_base('rpop', cmd)

    def zaddnr(self, cmd):
        self._redis_executer.zadd(cmd.key, **cmd.members_and_scores)

    def zscore(self, cmd):
        ret = self._redis_executer.zscore(cmd.key, cmd.member)
        try:
            return int(ret)
        except TypeError:
            return None

    def zrevrangebyscore(self, cmd):
        # redis> ZREVRANGEBYSCORE myzset +inf -inf
        # 1) "three"
        # 2) "two"
        # 3) "one"
        # redis> ZREVRANGEBYSCORE myzset 2 1
        # 1) "two"
        # 2) "one"
        # redis> ZREVRANGEBYSCORE myzset 2 (1
        # 1) "two"
        # redis> ZREVRANGEBYSCORE myzset (2 (1
        def format_redis_range_str(x: int, default: str, marked_inclusive: bool):
            if x is None:
                return default
            else:
                s = str(x)
                if marked_inclusive:
                    return s
                else:
                    return '(' + s

        cmd_parts = [ 'ZREVRANGEBYSCORE',
                      cmd.key,
                      format_redis_range_str(cmd.max, 'inf', cmd.inclusive[0]),
                      format_redis_range_str(cmd.min, '-inf', cmd.inclusive[1]),
            ]

        if cmd.with_scores:
            cmd_parts.append('WITHSCORES')

        full_cmd_str = ' '.join(cmd_parts)
        res = self._redis_executer.execute_command(full_cmd_str)

        if not cmd.with_scores:
            return list(map(lambda x: x.decode('utf-8'), res))
        else:
            res_byte_pairs = grouper(res, 2)
            return map(lambda x_y: ( int(x_y[1]), x_y[0].decode("utf-8") ), res_byte_pairs)


    def zremnr(self, cmd):
        self._redis_executer.zrem(cmd.key, *cmd.members)


    def zincrbynr(self, cmd):
        self._redis_executer.zincrby(cmd.key, cmd.member, cmd.amount)


    def transform_exception(f):
        @wraps(f)
        def wrapper(self, cmd):
            try:
                return f(self, cmd)
            except redis.exceptions.ResponseError as e:
                if e.args[0] == 'WRONGTYPE Operation against a key holding the wrong kind of value':
                    raise RedisKeyTypeError('Existing value has wrong type.')
                else:
                    raise

        return wrapper

    @transform_exception
    def __call__(self, cmd):
        # TODO: Make sure this is efficient and generally okay.

        if isinstance(cmd, Del):
            self.del_(cmd)
        else:
            return getattr(self, cmd.__class__.__name__.lower())(cmd)


def run_tests(deps_provider):
    ex = Executer(host='127.0.0.1', port=32768)
    from seneca.engine.util import return_exception_tuple
    import doctest, sys

    return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})
